/*
 *  Copyright (C) 2002-2007  The DOSBox Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* $Id: sdlmain.cpp,v 1.134 2007/08/26 18:03:25 qbix79 Exp $ */

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <stdarg.h>
#include <sys/types.h>
#ifdef WIN32
#include <signal.h>
#endif

#include "SDL.h"
#include "SDL_ttf.h"    // Used for OSD (GP2X)
#include <SDL_gp2x.h>	// GP2X Needed for direct touchscreen access
#include <fcntl.h>      // GP2X Need for MMU hack

#include "dosbox.h"
#include "video.h"
#include "mouse.h"
#include "pic.h"
#include "timer.h"
#include "setup.h"
#include "support.h"
#include "debug.h"
#include "mapper.h"
#include "vga.h"
#include "keyboard.h"
#include "programs.h"

//#define DISABLE_JOYSTICK

#if C_OPENGL
#include "SDL_opengl.h"

#ifndef APIENTRY
#define APIENTRY
#endif
#ifndef APIENTRYP
#define APIENTRYP APIENTRY *
#endif

#ifdef __WIN32__
#define NVIDIA_PixelDataRange 1

#ifndef WGL_NV_allocate_memory
#define WGL_NV_allocate_memory 1
typedef void * (APIENTRY * PFNWGLALLOCATEMEMORYNVPROC) (int size, float readfreq, float writefreq, float priority);
typedef void (APIENTRY * PFNWGLFREEMEMORYNVPROC) (void *pointer);
#endif

PFNWGLALLOCATEMEMORYNVPROC db_glAllocateMemoryNV = NULL;
PFNWGLFREEMEMORYNVPROC db_glFreeMemoryNV = NULL;

#else

#endif

#if defined(NVIDIA_PixelDataRange)

#ifndef GL_NV_pixel_data_range
#define GL_NV_pixel_data_range 1
#define GL_WRITE_PIXEL_DATA_RANGE_NV      0x8878
typedef void (APIENTRYP PFNGLPIXELDATARANGENVPROC) (GLenum target, GLsizei length, GLvoid *pointer);
typedef void (APIENTRYP PFNGLFLUSHPIXELDATARANGENVPROC) (GLenum target);
#endif

PFNGLPIXELDATARANGENVPROC glPixelDataRangeNV = NULL;

#endif

#endif //C_OPENGL

#if !(ENVIRON_INCLUDED)
extern char** environ;
#endif

#ifdef WIN32
#ifndef WIN32_LEAN_AND_MEAN
#define WIN32_LEAN_AND_MEAN
#endif
#include <windows.h>
#if (HAVE_DDRAW_H)
#include <ddraw.h>
struct private_hwdata {
	LPDIRECTDRAWSURFACE3 dd_surface;
	LPDIRECTDRAWSURFACE3 dd_writebuf;
};
#endif

#define STDOUT_FILE	TEXT("stdout.txt")
#define STDERR_FILE	TEXT("stderr.txt")
#define DEFAULT_CONFIG_FILE "/dosbox.conf"
#elif defined(MACOSX)
#define DEFAULT_CONFIG_FILE "/Library/Preferences/DOSBox Preferences"
#else /*linux freebsd*/
#define DEFAULT_CONFIG_FILE "/.dosboxrc"
#endif

#if C_SET_PRIORITY
#include <sys/resource.h>
#define PRIO_TOTAL (PRIO_MAX-PRIO_MIN)
#endif

#ifdef OS2
#define INCL_DOS
#define INCL_WIN
#include <os2.h>
#endif

void GP2X_RefreshScreen( void );

enum SCREEN_TYPES	{
	SCREEN_SURFACE,
	SCREEN_SURFACE_DDRAW,
	SCREEN_OVERLAY,
	SCREEN_OPENGL
};

enum PRIORITY_LEVELS {
	PRIORITY_LEVEL_PAUSE,
	PRIORITY_LEVEL_LOWEST,
	PRIORITY_LEVEL_LOWER,
	PRIORITY_LEVEL_NORMAL,
	PRIORITY_LEVEL_HIGHER,
	PRIORITY_LEVEL_HIGHEST
};

// Note: these defines can go away once a runtime solution is made for the joystick
/* depending on what SDL libs you have, some keys may be different to whats set here. For example
the L and R shoulder buttons may be swapped, if so swap them in the defines below.*/
#define GP2X_BUTTON_UP              (0)
#define GP2X_BUTTON_DOWN            (4)
#define GP2X_BUTTON_LEFT            (2)
#define GP2X_BUTTON_RIGHT           (6)
#define GP2X_BUTTON_UPLEFT          (1)
#define GP2X_BUTTON_UPRIGHT         (7)
#define GP2X_BUTTON_DOWNLEFT        (3)
#define GP2X_BUTTON_DOWNRIGHT       (5)
#define GP2X_BUTTON_CLICK           (18)
#define GP2X_BUTTON_A               (12)
#define GP2X_BUTTON_B               (13)
#define GP2X_BUTTON_X               (15)
#define GP2X_BUTTON_Y               (14)
#define GP2X_BUTTON_L               (11)
#define GP2X_BUTTON_R               (10)
#define GP2X_BUTTON_START           (8)
#define GP2X_BUTTON_SELECT          (9)
#define GP2X_BUTTON_VOLUP           (16)
#define GP2X_BUTTON_VOLDOWN         (17)

bool GP2X_ButtonPressed[19];
bool saveconfig = false;

struct SDL_Block {
	bool active;							//If this isn't set don't draw
	bool updating;
	struct {
		Bit32u width;
		Bit32u height;
		Bit32u bpp;
		Bitu flags;
		double scalex,scaley;
		GFX_CallBack_t callback;
	} draw;
	bool wait_on_error;
	struct {
		struct {
			Bit16u width, height;
			bool fixed;
		} full;
		struct {
			Bit16u width, height;
		} window;
		Bit8u bpp;
		bool fullscreen;
		bool doublebuf;
		SCREEN_TYPES type;
		SCREEN_TYPES want_type;
	} desktop;
#if C_OPENGL
	struct {
		Bitu pitch;
		void * framebuf;
		GLuint texture;
		GLuint displaylist;
		GLint max_texsize;
		bool bilinear;
		bool packed_pixel;
		bool paletted_texture;
#if defined(NVIDIA_PixelDataRange)
		bool pixel_data_range;
#endif
	} opengl;
#endif
	struct {
		SDL_Surface * surface;
		SDL_Surface * buffer;
#if (HAVE_DDRAW_H) && defined(WIN32)
		RECT rect;
#endif
	} blit;
	struct {
		PRIORITY_LEVELS focus;
		PRIORITY_LEVELS nofocus;
	} priority;
	SDL_Rect clip;
	SDL_Surface * surface;
	SDL_Overlay * overlay;
	SDL_cond *cond;
	struct {
		bool autolock;
		bool autoenable;
		bool requestlock;
		bool locked;
		Bitu sensitivity;
	} mouse;
	SDL_Rect updateRects[1024];
	Bitu num_joysticks;
#if defined (WIN32)
	bool using_windib;
#endif
	TTF_Font *font;     // GP2X OSD Begin
	int fontsize;       // GP2X OSD Font Size
	char title[200];
	Uint32 fpstime;
	int fps, curfps;    // GP2X OSD Enf

};

static SDL_Block sdl;

extern const char* RunningProgram;
extern bool CPU_CycleAutoAdjust;
//Globals for keyboard initialisation
bool startup_state_numlock=false;
bool startup_state_capslock=false;


// GP2X OSD Color Variables
int fontcolorset = 0;
int fontcolor_front;
int fontcolor_back;
int fontcolor_frontsel;
int fontcolor_backsel;

static void SaveConfigSetting(int cycles, int frameskip)
{
    bool clrft  = false;
    bool clrbk  = false;
    bool clrfts = false;
    bool clrbks = false;
    char line[100];
    char* ptr_line = &line[0];

    FILE* fConf = fopen( "dosbox.conf", "r" );
    FILE* fTemp = fopen( "temp.conf", "w" );
    if(fConf)
    {
        while( !feof(fConf) )
        {
            if( fgets( ptr_line, sizeof(line), fConf ) != NULL )
            {
                if( strstr( ptr_line, "cycles=" ) != NULL && cycles > -1 )
                {
                    fprintf( fTemp, "cycles=%d\n", cycles );
                }
                else if( strstr( ptr_line, "frameskip=" ) != NULL && frameskip > -1 )
                {
                    fprintf( fTemp, "frameskip=%d\n", frameskip );
                }
                else if( strstr( ptr_line, "fontcolor_front=" ) != NULL )
                {
                    fprintf( fTemp, "fontcolor_front=%d\n", fontcolor_front );
                    clrft = true;
                }
                else if( strstr( ptr_line, "fontcolor_back=" ) != NULL )
                {
                    fprintf( fTemp, "fontcolor_back=%d\n", fontcolor_back );
                    clrbk = true;
                }
                else if( strstr( ptr_line, "fontcolor_frontsel=" ) != NULL )
                {
                    fprintf( fTemp, "fontcolor_frontsel=%d\n", fontcolor_frontsel );
                    clrfts = true;
                }
                else if( strstr( ptr_line, "fontcolor_backsel=" ) != NULL )
                {
                    fprintf( fTemp, "fontcolor_backsel=%d\n", fontcolor_backsel );
                    clrbks = true;
                }
                else if( strstr( ptr_line, "[dosbox]" ) != NULL )
                {
                    if(!clrft && !clrbk && !clrfts && !clrbks)
                    {
                        fprintf( fTemp, "fontcolor_front=%d\nfontcolor_back=%d\nfontcolor_frontsel=%d\nfontcolor_backsel=%d\n", fontcolor_front
                                                                                                                              , fontcolor_back
                                                                                                                              , fontcolor_frontsel
                                                                                                                              , fontcolor_backsel );
                    }
                    fprintf( fTemp, "%s", line );
                }
                else
                {
                    fprintf( fTemp, "%s", line );
                }
            }
        }

        fclose(fConf);
        fclose(fTemp);
    }

    remove( "dosbox.conf" );
    rename( "temp.conf", "dosbox.conf" );
}

void GFX_SetTitle(Bit32s cycles,Bits frameskip,bool paused){
    bool refresh;
    static char cycles_label[3] = " C";
	char title[200]={0};
	static Bit32s internal_cycles=0;
	static Bits internal_frameskip=0;
	static Bit32s internal_cycles_prev=0;
	static Bits internal_frameskip_prev=0;
	static Bit32s internal_cycles_saved=-1;
	static Bits internal_frameskip_saved=-1;

    refresh = false;

	if(cycles != -1) internal_cycles = cycles;
	if(frameskip != -1) internal_frameskip = frameskip;
	if(CPU_CycleAutoAdjust) {
		if (internal_cycles>=100)
			sprintf(title,"DOSBox %s, Cpu Cycles:      max, Frameskip %2d, Program: %8s",VERSION,internal_frameskip,RunningProgram);
		else
			sprintf(title,"DOSBox %s, Cpu Cycles:   [%3d%%], Frameskip %2d, Program: %8s",VERSION,internal_cycles,internal_frameskip,RunningProgram);
	} else {
		sprintf(title,"DOSBox %s, Cpu Cycles: %8d, Frameskip %2d, Program: %8s",VERSION,internal_cycles,internal_frameskip,RunningProgram);
	}

	if(paused) strcat(title," PAUSED");
	SDL_WM_SetCaption(title,VERSION);

	// on-screen title BEGIN (GP2X)
        // Set the parameters to the current ones (these should be the ones from the current dosbox.conf)
    if( internal_cycles_saved == -1 && internal_cycles > 0 ) internal_cycles_saved = internal_cycles;
	if( internal_frameskip_saved == -1 && internal_frameskip > -1 ) internal_frameskip_saved = internal_frameskip;

	if( internal_cycles != internal_cycles_prev || internal_frameskip != internal_frameskip_prev ) {
        refresh = true;
	}

    // The user has signaled to save the config the current setting are now the ones in file
    if( saveconfig )
    {
        if( CPU_CycleAutoAdjust && internal_cycles<100 )
        {
            // Dont need cycles, its set to auto or max
            SaveConfigSetting( -1, internal_frameskip );
        }
        else {
            SaveConfigSetting( internal_cycles, internal_frameskip );
        }

        internal_cycles_saved = internal_cycles;
        internal_frameskip_saved = internal_frameskip;
        saveconfig = false;
    }

        // If the parameters change then set the star to notify the user (* means not saved)
	if( internal_cycles == internal_cycles_saved && internal_frameskip == internal_frameskip_saved ) {
	    if( cycles_label[0] != ' ' ) refresh = true;
	    cycles_label[0] = ' ';
	}
	else {
	    if( cycles_label[0] != '*' ) refresh = true;
	    cycles_label[0] = '*';
	}

	if(CPU_CycleAutoAdjust) {
		if (internal_cycles>=100) {
			sprintf(title,"%s=max FS=%2d",cycles_label,internal_frameskip);
		}
		else {
			sprintf(title,"%s=[%d%%] FS=%2d",cycles_label,internal_cycles,internal_frameskip);
		}
	} else {
		sprintf(title,"%s=%d FS=%2d",cycles_label,internal_cycles,internal_frameskip);
	}
	strcpy(sdl.title, title);

    internal_cycles_prev = internal_cycles;
    internal_frameskip_prev = internal_frameskip;

    if( refresh ) GP2X_RefreshScreen();
    // on-screen title END (GP2X)
}

static void PauseDOSBox(bool pressed) {
	if (!pressed)
		return;
	GFX_SetTitle(-1,-1,true);
	bool paused = true;
	KEYBOARD_ClrBuffer();
	SDL_Delay(500);
	SDL_Event event;
	while (SDL_PollEvent(&event)) {
		// flush event queue.
	}
	while (paused) {
		SDL_WaitEvent(&event);    // since we're not polling, cpu usage drops to 0.
		switch (event.type) {
		case SDL_QUIT: throw(0); break;
		case SDL_KEYDOWN:   // Must use Pause/Break Key to resume.
		case SDL_KEYUP:
			if(event.key.keysym.sym==SDLK_PAUSE){
				paused=false;
				GFX_SetTitle(-1,-1,false);
				break;
			}
		}
	}
}

void GP2X_LoadFont( int fontsize )
{
    if (sdl.font) TTF_CloseFont(sdl.font);

    sdl.font=TTF_OpenFont("dbfont.ttf", fontsize );
    if (!sdl.font)  // Try again!
        sdl.font=TTF_OpenFont("../dbfont.ttf", fontsize );

    if (!sdl.font)  // Didnt find the font
    {
        LOG_MSG("Could not load dbfont.ttf, GP2X debug disabled (%s)", TTF_GetError());
    }
    else
    {
        if (fontsize > 14 ) {
            TTF_SetFontStyle(sdl.font, TTF_STYLE_BOLD );
        }
        else {
            TTF_SetFontStyle(sdl.font, TTF_STYLE_NORMAL );
        }
    }
}

#define BASIC_COLORS      16
#define VIRT_KEYBOARD      2
#define VIRT_KEYBOARD_ROW  9
#define VIRT_KEYBOARD_COL 10

// Normal Keyboard Row
#define VIRT_ROW_SHIFTBACK 5
#define VIRT_ROW_LALTCTRL  6
#define VIRT_ROW_RALTCTRL  7

// Special Keyboard Row
#define VIRT_ROW_INSHOM    0
#define VIRT_ROW_DELEND    1
#define VIRT_ROW_PGUPGD    2
#define VIRT_ROW_ESCENTER  3
#define VIRT_ROW_TABPAUSE  4
#define VIRT_ROW_FUNCLABEL 5
#define VIRT_ROW_FUNC1_10  6
#define VIRT_ROW_FUNC1112  7
// Both Keyboard Row
#define VIRT_ROW_SPECIAL   8

//    0   1   2   3   4   5   6   7   8   9
char gp2x_keyboard[VIRT_KEYBOARD][VIRT_KEYBOARD_ROW][VIRT_KEYBOARD_COL] =
{
 { { '1','2','3','4','5','6','7','8','9','0' },
   { 'a','b','c','d','e','f','g','h','i','j' },
   { 'k','l','m','n','o','p','q','r','s','t' },
   { 'u','v','w','x','y','z',' ',' ',' ',0x27 },
   { '`','-','=','[',']',';',',','.','/','\\' },
   { 'S','H','F','T',' ','B','A','C','K',' ' },
   { 'L','A','L','T',' ','L','C','T','L',' ' },
   { 'R','A','L','T',' ','R','C','T','L',' ' },
   { 'S','P','E','C','A','L','K','E','Y','S' } },
 { { 'I','N','S',' ',' ','H','O','M','E',' ' },
   { 'D','E','L',' ',' ','E','N','D',' ',' ' },
   { 'P','G','U','P',' ','P','G','D','W','N' },
   { 'E','S','C',' ',' ','E','N','T','E','R' },
   { 'T','A','B',' ',' ','P','A','U','S','E' },
   { 'F','U','N','C',' ',' ','K','E','Y','S' },
   { '1','2','3','4','5','6','7','8','9','0' },
   { '1','1',' ',' ',' ','1','2',' ',' ',' ' },
   { 'N','O','R','M','A','L','K','E','Y','S' } }
};

int gp2x_cur_kbd = 0;
int gp2x_col_key = 0;
int gp2x_row_key = 0;
                                        //   r    g    b
SDL_Color basic_colors[BASIC_COLORS] = { { 255, 255, 255, 0 },    //  0 = white
                                         {   0,   0,   0, 0 },    //  1 = black
                                         { 255,   0,   0, 0 },    //  2 = red
                                         {   0,   0, 255, 0 },    //  3 = blue
                                         {   0, 255,   0, 0 },    //  4 = green
                                         { 255, 255,   0, 0 },    //  5 = yellow
                                         {   0, 255, 255, 0 },    //  6 = cyan
                                         { 255,   0, 127, 0 },    //  7 = pink
                                         { 255, 127,   0, 0 },    //  8 = orange
                                         {   0, 127, 127, 0 },    //  9 = dcyan
                                         {  63,  63,  63, 0 },    // 10 = grey
                                         { 191, 191, 191, 0 },    // 11 = ltgrey
                                         {   0,   0, 127, 0 },    // 12 = dkblue
                                         { 127,   0,   0, 0 },    // 13 = dkred
                                         {   0, 127,   0, 0 },    // 14 = dkgreen
                                         { 100,   0,   0, 0 } };  // 15 = use for transparent

bool gp2x_shift          = false;
bool gp2x_lalt           = false;
bool gp2x_lctrl          = false;
bool gp2x_ralt           = false;
bool gp2x_rctrl          = false;
bool gp2x_low_halfrow    = false;
bool gp2x_up_halfrow     = false;
bool gp2x_full_row       = false;

#define OSD_STATE_NOBUFFER 1
#define OSD_STATE_BUFFER   2
int gp2x_osd_bufferstate = OSD_STATE_NOBUFFER; // Use defined states above
bool gp2x_osd_debug     = false;
bool gp2x_osd_virtkey   = false;
bool gp2x_virtkey_press = false;
bool gp2x_vk_firstrun   = false;

void GP2X_OSD_GFX_Change( void )
{
    if( gp2x_osd_debug == false && gp2x_osd_virtkey == false && gp2x_osd_bufferstate == OSD_STATE_BUFFER )
    {
        SDL_BlitSurface(sdl.blit.surface, NULL,sdl.surface, NULL );
        gp2x_osd_bufferstate = OSD_STATE_NOBUFFER;
    }
    else if( (gp2x_osd_debug == true || gp2x_osd_virtkey == true) && gp2x_osd_bufferstate == OSD_STATE_NOBUFFER )
    {
        SDL_BlitSurface(sdl.surface, NULL,sdl.blit.surface, NULL );
        gp2x_osd_bufferstate = OSD_STATE_BUFFER;
    }
}

void GP2X_RefreshScreen( void )
{
    if( sdl.updating == false )
    {
        GP2X_OSD_GFX_Change();
        sdl.updating=true;
        GFX_EndUpdate( NULL );
    }
}

void GP2X_Debug_Toggle(bool pressed)
{
	if (!pressed) return;
	gp2x_osd_debug=!gp2x_osd_debug;
	GP2X_RefreshScreen();
}

void GP2X_VirtualKeyboard_Toggle( bool pressed )
{
	if (!pressed) return;
	gp2x_osd_virtkey=!gp2x_osd_virtkey;
	if( gp2x_osd_virtkey ) gp2x_vk_firstrun = true;
	GP2X_RefreshScreen();
}

int GP2X_VirtualKeyboard( SDL_Event* event )
{
    int intReturnKey = 0;
    bool update_screen = false;
    int GP2X_button = event->jbutton.button;
    bool button_state = (event->type == SDL_JOYBUTTONDOWN) ? true : false;

    // Move to the new position
    if( button_state )
    {
        switch( GP2X_button )
        {
            case GP2X_BUTTON_DOWN:
                gp2x_row_key++;
                if( gp2x_row_key >= VIRT_KEYBOARD_ROW ) gp2x_row_key = 0;

                if( gp2x_cur_kbd == 1 && gp2x_row_key == VIRT_ROW_FUNCLABEL ) gp2x_row_key++; // Dummy labels

                if( gp2x_row_key == VIRT_ROW_SPECIAL )
                {
                    gp2x_low_halfrow    = false;
                    gp2x_up_halfrow     = false;
                    gp2x_full_row       = true;
                }
                else
                {
                    gp2x_low_halfrow    = false;
                    gp2x_up_halfrow     = false;
                    gp2x_full_row       = false;
                }

                if( (gp2x_cur_kbd == 0 && ( gp2x_row_key == VIRT_ROW_SHIFTBACK || gp2x_row_key == VIRT_ROW_LALTCTRL || gp2x_row_key == VIRT_ROW_RALTCTRL )) ||
                    (gp2x_cur_kbd == 1 && ( gp2x_row_key == VIRT_ROW_INSHOM    || gp2x_row_key == VIRT_ROW_DELEND   ||
                                            gp2x_row_key == VIRT_ROW_PGUPGD    || gp2x_row_key == VIRT_ROW_ESCENTER || gp2x_row_key == VIRT_ROW_TABPAUSE ||
                                            gp2x_row_key == VIRT_ROW_FUNCLABEL || gp2x_row_key == VIRT_ROW_FUNC1112 )) )
                {
                    if( gp2x_col_key < VIRT_KEYBOARD_COL/2 )
                    {
                        gp2x_col_key = 0;
                        gp2x_low_halfrow    = true;
                        gp2x_up_halfrow     = false;
                        gp2x_full_row       = false;
                    }
                    else
                    {
                        gp2x_col_key = VIRT_KEYBOARD_COL/2;
                        gp2x_low_halfrow    = false;
                        gp2x_up_halfrow     = true;
                        gp2x_full_row       = false;
                    }
                }
                update_screen = true;
                break;
            case GP2X_BUTTON_UP:
                gp2x_row_key--;
                if( gp2x_row_key < 0 ) gp2x_row_key = VIRT_KEYBOARD_ROW-1;

                if( gp2x_cur_kbd == 1 && gp2x_row_key == VIRT_ROW_FUNCLABEL ) gp2x_row_key--; // Dummy labels

                if( gp2x_row_key == VIRT_ROW_SPECIAL )
                {
                    gp2x_low_halfrow    = false;
                    gp2x_up_halfrow     = false;
                    gp2x_full_row       = true;
                }
                else
                {
                    gp2x_low_halfrow    = false;
                    gp2x_up_halfrow     = false;
                    gp2x_full_row       = false;
                }

                if( (gp2x_cur_kbd == 0 && ( gp2x_row_key == VIRT_ROW_SHIFTBACK || gp2x_row_key == VIRT_ROW_LALTCTRL || gp2x_row_key == VIRT_ROW_RALTCTRL )) ||
                    (gp2x_cur_kbd == 1 && ( gp2x_row_key == VIRT_ROW_INSHOM    || gp2x_row_key == VIRT_ROW_DELEND   ||
                                            gp2x_row_key == VIRT_ROW_PGUPGD    || gp2x_row_key == VIRT_ROW_ESCENTER || gp2x_row_key == VIRT_ROW_TABPAUSE ||
                                            gp2x_row_key == VIRT_ROW_FUNCLABEL || gp2x_row_key == VIRT_ROW_FUNC1112 )) )
                {
                    if( gp2x_col_key < VIRT_KEYBOARD_COL/2 )
                    {
                        gp2x_col_key = 0;
                        gp2x_low_halfrow    = true;
                        gp2x_up_halfrow     = false;
                        gp2x_full_row       = false;
                    }
                    else
                    {
                        gp2x_col_key = VIRT_KEYBOARD_COL/2;
                        gp2x_low_halfrow    = false;
                        gp2x_up_halfrow     = true;
                        gp2x_full_row       = false;
                    }
                }
                update_screen = true;
                break;
            case GP2X_BUTTON_RIGHT:
                gp2x_col_key++;
                if( gp2x_col_key >= VIRT_KEYBOARD_COL ) gp2x_col_key = 0;

                if( (gp2x_cur_kbd == 0 && ( gp2x_row_key == VIRT_ROW_SHIFTBACK || gp2x_row_key == VIRT_ROW_LALTCTRL || gp2x_row_key == VIRT_ROW_RALTCTRL )) ||
                    (gp2x_cur_kbd == 1 && ( gp2x_row_key == VIRT_ROW_INSHOM    || gp2x_row_key == VIRT_ROW_DELEND   ||
                                            gp2x_row_key == VIRT_ROW_PGUPGD    || gp2x_row_key == VIRT_ROW_ESCENTER || gp2x_row_key == VIRT_ROW_TABPAUSE ||
                                            gp2x_row_key == VIRT_ROW_FUNCLABEL || gp2x_row_key == VIRT_ROW_FUNC1112 )) )
                {
                    if( gp2x_col_key < VIRT_KEYBOARD_COL/2 )
                    {
                        gp2x_col_key = VIRT_KEYBOARD_COL/2;
                        gp2x_low_halfrow    = false;
                        gp2x_up_halfrow     = true;
                        gp2x_full_row       = false;
                    }
                    else
                    {
                        gp2x_col_key = 0;
                        gp2x_low_halfrow    = true;
                        gp2x_up_halfrow     = false;
                        gp2x_full_row       = false;
                    }
                }
                update_screen = true;
                break;
            case GP2X_BUTTON_LEFT:
                gp2x_col_key--;
                if( gp2x_col_key < 0 ) gp2x_col_key = VIRT_KEYBOARD_COL-1;

                if( (gp2x_cur_kbd == 0 && ( gp2x_row_key == VIRT_ROW_SHIFTBACK || gp2x_row_key == VIRT_ROW_LALTCTRL || gp2x_row_key == VIRT_ROW_RALTCTRL )) ||
                    (gp2x_cur_kbd == 1 && ( gp2x_row_key == VIRT_ROW_INSHOM    || gp2x_row_key == VIRT_ROW_DELEND   ||
                                            gp2x_row_key == VIRT_ROW_PGUPGD    || gp2x_row_key == VIRT_ROW_ESCENTER || gp2x_row_key == VIRT_ROW_TABPAUSE ||
                                            gp2x_row_key == VIRT_ROW_FUNCLABEL || gp2x_row_key == VIRT_ROW_FUNC1112 )) )
                {
                    if( gp2x_col_key < 0  )
                    {
                        gp2x_col_key = VIRT_KEYBOARD_COL/2;
                        gp2x_low_halfrow    = false;
                        gp2x_up_halfrow     = true;
                        gp2x_full_row       = false;
                    }
                    else
                    {
                        gp2x_col_key = 0;
                        gp2x_low_halfrow    = true;
                        gp2x_up_halfrow     = false;
                        gp2x_full_row       = false;
                    }
                }
                update_screen = true;
                break;
            case GP2X_BUTTON_VOLUP:
                if( fontcolorset == 0 )
                {
                    fontcolor_front++;
                }
                else
                {
                    fontcolor_frontsel++;
                }
                if( fontcolor_front    >= BASIC_COLORS ) fontcolor_front = 0;
                if( fontcolor_frontsel >= BASIC_COLORS ) fontcolor_frontsel = 0;
                update_screen = true;
                break;
            case GP2X_BUTTON_VOLDOWN:
                if( fontcolorset == 0 )
                {
                    fontcolor_back++;
                }
                else
                {
                   fontcolor_backsel++;
                }
                if( fontcolor_back    >= BASIC_COLORS ) fontcolor_back = 0;
                if( fontcolor_backsel >= BASIC_COLORS ) fontcolor_backsel = 0;
                update_screen = true;
                break;
            case GP2X_BUTTON_SELECT:
                if( fontcolorset == 0 )
                {
                    fontcolorset = 1;
                }
                else
                {
                    fontcolorset = 0;
                }
                break;
        }
    }

    event->key.keysym.sym = (SDLKey)NULL;
    switch( GP2X_button )
    {
        case GP2X_BUTTON_START:
            if(!button_state && !gp2x_vk_firstrun) GP2X_VirtualKeyboard_Toggle( true );  // This should close the osd upon releasing the start button
            if( button_state || gp2x_vk_firstrun ) gp2x_vk_firstrun = false;
            break;
        case GP2X_BUTTON_L:
        case GP2X_BUTTON_R:
            if( button_state )
            {
                gp2x_row_key = 0;
                gp2x_col_key = 0;

                if( gp2x_cur_kbd == 0)
                {
                    gp2x_cur_kbd = 1;
                    gp2x_low_halfrow    = true;
                    gp2x_up_halfrow     = false;
                    gp2x_full_row       = false;
                }
                else
                {
                    gp2x_cur_kbd = 0;
                    gp2x_low_halfrow    = false;
                    gp2x_up_halfrow     = false;
                    gp2x_full_row       = false;
                }
                update_screen = true;
            }
            break;
        case GP2X_BUTTON_X:
            event->key.keysym.sym = SDLK_BACKSPACE;
            break;
        case GP2X_BUTTON_Y:
            event->key.keysym.sym = SDLK_SPACE;
            break;
        case GP2X_BUTTON_A:
            event->key.keysym.sym = SDLK_RETURN;
            break;
        case GP2X_BUTTON_B:
            if( gp2x_cur_kbd == 0 )
            {
                switch( gp2x_row_key )
                {
                    case VIRT_ROW_SPECIAL:
                        if( button_state )
                        {
                            gp2x_cur_kbd = 1;

                            gp2x_row_key = 0;
                            gp2x_col_key = 0;
                            gp2x_low_halfrow    = true;
                            gp2x_up_halfrow     = false;
                            gp2x_full_row       = false;
                            update_screen = true;
                        }
                        break;
                    case VIRT_ROW_SHIFTBACK:
                        if( gp2x_col_key < VIRT_KEYBOARD_COL/2 )
                        {
                            if( button_state )
                            {
                                if( gp2x_shift == false )
                                {
                                    button_state = true;
                                    gp2x_shift   = true;
                                    gp2x_keyboard[gp2x_cur_kbd][VIRT_ROW_SHIFTBACK][(VIRT_KEYBOARD_COL/2)-1] = 'X';
                                }
                                else
                                {
                                    button_state = false;
                                    gp2x_shift   = false;
                                    gp2x_keyboard[gp2x_cur_kbd][VIRT_ROW_SHIFTBACK][(VIRT_KEYBOARD_COL/2)-1] = ' ';
                                }
                                update_screen = true;
                                event->key.keysym.sym = SDLK_LSHIFT;
                            }
                        }
                        else if( gp2x_col_key >= VIRT_KEYBOARD_COL/2 )
                        {
                            event->key.keysym.sym = SDLK_BACKSPACE;
                        }
                        break;
                    case VIRT_ROW_LALTCTRL:
                        if( gp2x_col_key < VIRT_KEYBOARD_COL/2 )
                        {
                            if( button_state )
                            {
                                if( gp2x_lalt == false )
                                {
                                    button_state = true;
                                    gp2x_lalt    = true;
                                    gp2x_keyboard[gp2x_cur_kbd][VIRT_ROW_LALTCTRL][(VIRT_KEYBOARD_COL/2)-1] = 'X';
                                }
                                else
                                {
                                    button_state = false;
                                    gp2x_lalt    = false;
                                    gp2x_keyboard[gp2x_cur_kbd][VIRT_ROW_LALTCTRL][(VIRT_KEYBOARD_COL/2)-1] = ' ';
                                }
                                update_screen = true;
                                event->key.keysym.sym = SDLK_LALT;
                            }
                        }
                        else if( gp2x_col_key >= VIRT_KEYBOARD_COL/2 )
                        {
                            if( button_state )
                            {
                                if( gp2x_lctrl == false )
                                {
                                    button_state = true;
                                    gp2x_lctrl   = true;
                                    gp2x_keyboard[gp2x_cur_kbd][VIRT_ROW_LALTCTRL][VIRT_KEYBOARD_COL-1] = 'X';
                                }
                                else
                                {
                                    button_state = false;
                                    gp2x_lctrl   = false;
                                    gp2x_keyboard[gp2x_cur_kbd][VIRT_ROW_LALTCTRL][VIRT_KEYBOARD_COL-1] = ' ';
                                }
                                update_screen = true;
                                event->key.keysym.sym = SDLK_LCTRL;
                            }
                        }
                        break;
                    case VIRT_ROW_RALTCTRL:
                        if( gp2x_col_key < VIRT_KEYBOARD_COL/2 )
                        {
                            if( button_state )
                            {
                                if( gp2x_ralt == false )
                                {
                                    button_state = true;
                                    gp2x_ralt    = true;
                                    gp2x_keyboard[gp2x_cur_kbd][VIRT_ROW_RALTCTRL][(VIRT_KEYBOARD_COL/2)-1] = 'X';
                                }
                                else
                                {
                                    button_state = false;
                                    gp2x_ralt    = false;
                                    gp2x_keyboard[gp2x_cur_kbd][VIRT_ROW_RALTCTRL][(VIRT_KEYBOARD_COL/2)-1] = ' ';
                                }
                                update_screen = true;
                                event->key.keysym.sym = SDLK_RALT;
                            }
                        }
                        else if( gp2x_col_key >= VIRT_KEYBOARD_COL/2 )
                        {
                            if( button_state )
                            {
                                if( gp2x_rctrl == false )
                                {
                                    button_state = true;
                                    gp2x_rctrl   = true;
                                    gp2x_keyboard[gp2x_cur_kbd][VIRT_ROW_RALTCTRL][VIRT_KEYBOARD_COL-1] = 'X';
                                }
                                else
                                {
                                    button_state = false;
                                    gp2x_rctrl   = false;
                                    gp2x_keyboard[gp2x_cur_kbd][VIRT_ROW_RALTCTRL][VIRT_KEYBOARD_COL-1] = ' ';
                                }
                                update_screen = true;
                                event->key.keysym.sym = SDLK_RCTRL;
                            }
                        }
                        break;
                    default:
                        event->key.keysym.sym = (SDLKey)gp2x_keyboard[gp2x_cur_kbd][gp2x_row_key][gp2x_col_key];
                        break;
                }
            }
            else if( gp2x_cur_kbd == 1 )
            {
                switch( gp2x_row_key )
                {
                    case VIRT_ROW_SPECIAL:
                        if( button_state )
                        {
                            gp2x_cur_kbd = 0;

                            gp2x_row_key = 0;
                            gp2x_col_key = 0;
                            gp2x_low_halfrow    = false;
                            gp2x_up_halfrow     = false;
                            gp2x_full_row       = false;
                            update_screen = true;
                        }
                        break;
                    case VIRT_ROW_INSHOM:
                        if( gp2x_col_key < VIRT_KEYBOARD_COL/2 )
                        {
                            event->key.keysym.sym = SDLK_INSERT;
                        }
                        else if( gp2x_col_key >= VIRT_KEYBOARD_COL/2 )
                        {
                            event->key.keysym.sym = SDLK_HOME;
                        }
                        break;
                    case VIRT_ROW_DELEND:
                        if( gp2x_col_key < VIRT_KEYBOARD_COL/2 )
                        {
                            event->key.keysym.sym = SDLK_DELETE;
                        }
                        else if( gp2x_col_key >= VIRT_KEYBOARD_COL/2 )
                        {
                            event->key.keysym.sym = SDLK_END;
                        }
                        break;
                    case VIRT_ROW_PGUPGD:
                        if( gp2x_col_key < VIRT_KEYBOARD_COL/2 )
                        {
                            event->key.keysym.sym = SDLK_PAGEUP;
                        }
                        else if( gp2x_col_key >= VIRT_KEYBOARD_COL/2 )
                        {
                            event->key.keysym.sym = SDLK_PAGEDOWN;
                        }
                        break;
                    case VIRT_ROW_ESCENTER:
                        if( gp2x_col_key < VIRT_KEYBOARD_COL/2 )
                        {
                            event->key.keysym.sym = SDLK_ESCAPE;
                        }
                        else if( gp2x_col_key >= VIRT_KEYBOARD_COL/2 )
                        {
                            event->key.keysym.sym = SDLK_RETURN;
                        }
                        break;
                    case VIRT_ROW_TABPAUSE:
                        if( gp2x_col_key < VIRT_KEYBOARD_COL/2 )
                        {
                            event->key.keysym.sym = SDLK_TAB;
                        }
                        else if( gp2x_col_key >= VIRT_KEYBOARD_COL/2 )
                        {
                            event->key.keysym.sym = SDLK_PAUSE;
                        }
                        break;
                    case VIRT_ROW_FUNC1_10:
                        event->key.keysym.sym = (SDLKey)(SDLK_F1+gp2x_col_key);
                        break;
                    case VIRT_ROW_FUNC1112:
                        if( gp2x_col_key < VIRT_KEYBOARD_COL/2 )
                        {
                            event->key.keysym.sym = SDLK_F11;
                        }
                        else if( gp2x_col_key >= VIRT_KEYBOARD_COL/2 )
                        {
                            event->key.keysym.sym = SDLK_F12;
                        }
                        break;
                    default:
                        event->key.keysym.sym = (SDLKey)gp2x_keyboard[gp2x_cur_kbd][gp2x_row_key][gp2x_col_key];
                        break;
                }
            }
            break;
    }

    if( event->key.keysym.sym != (SDLKey)NULL )
    {
        if( button_state )
        {
            event->type      = SDL_KEYDOWN;
            event->key.state = SDL_PRESSED;
        }
        else
        {
            event->type      = SDL_KEYUP;
            event->key.state = SDL_RELEASED;
        }
        event->key.keysym.scancode   = 0;
        event->key.keysym.mod        = KMOD_NONE;
        event->key.keysym.unicode    = 0;
        intReturnKey = 1;
    }

    if( update_screen == true )
    {
        GP2X_RefreshScreen();
    }

    return intReturnKey;
}

// GP2X fullscreen/scale option
bool gp2x_mouse          = true;
bool button_state;
int gp2x_mouse_speed     = 0;
int gp2x_mouse_pixel     = 1;
int GP2X_MOUSE_SPEED     = -1;

bool gp2x_noclick        = false;
bool blnStickUp          = false;
bool blnStickDown        = false;
bool blnStickLeft        = false;
bool blnStickRight       = false;
bool blnStickUpRight     = false;
bool blnStickUpLeft      = false;
bool blnStickDownRight   = false;
bool blnStickDownLeft    = false;

// GP2X zoom feature
bool gp2xzoom;
int gp2xzoom_x, gp2xzoom_y;
int gp2xzoom_pct;
int gp2xzoom_w, gp2xzoom_h;

static void GP2X_ZoomReset() {
	gp2xzoom=false;
	gp2xzoom_pct=100;
	gp2xzoom_x=gp2xzoom_y=0;
	gp2xzoom_w=320;
	gp2xzoom_h=240;
}

static void GP2X_ZoomSet() {
	SDL_GP2X_MiniDisplay(0, 0); // must reset first

	SDL_Rect displayarea;
	displayarea.x=gp2xzoom_x;
	displayarea.y=gp2xzoom_y;
	displayarea.w=gp2xzoom_w;
	displayarea.h=gp2xzoom_h;
	SDL_GP2X_Display(&displayarea);
}

static void GP2X_ZoomPct() {
	gp2xzoom_w=(int)((float)SDL_GetVideoSurface()->w*((float)gp2xzoom_pct/100));
	gp2xzoom_h=(int)((float)SDL_GetVideoSurface()->h*((float)gp2xzoom_pct/100));

	if (gp2xzoom_w>SDL_GetVideoSurface()->w) gp2xzoom_w=SDL_GetVideoSurface()->w;
	if (gp2xzoom_h>SDL_GetVideoSurface()->h) gp2xzoom_h=SDL_GetVideoSurface()->h;

	if (gp2xzoom_x>SDL_GetVideoSurface()->w-gp2xzoom_w) gp2xzoom_x=SDL_GetVideoSurface()->w-gp2xzoom_w;
	if (gp2xzoom_y>SDL_GetVideoSurface()->h-gp2xzoom_h) gp2xzoom_y=SDL_GetVideoSurface()->h-gp2xzoom_h;

	GP2X_ZoomSet();
}

static void GP2X_Zoom(bool pressed) {
	if (!pressed) return;

    LOG_MSG("GP2XZOOM: on");
    gp2xzoom_w=320;
    gp2xzoom_h=240;
    gp2xzoom_pct=100;
    GP2X_ZoomSet();
    gp2xzoom=true;
}
static void GP2X_NoZoom(bool pressed) {
	if (!pressed) return;

    if (SDL_GetVideoSurface()->w == 320 && SDL_GetVideoSurface()->h < 240) {
	    SDL_GP2X_MiniDisplay(0, 0);
	} else {
		SDL_Rect displayarea;
		displayarea.x=displayarea.y=0;
		displayarea.w=SDL_GetVideoSurface()->w;
		displayarea.h=SDL_GetVideoSurface()->h;
		LOG_MSG("GP2XZOOM: off");
		gp2xzoom_pct=100;
		SDL_GP2X_Display(&displayarea);
		gp2xzoom=false;
	}
}
static void GP2X_ZoomLeft(bool pressed) {
	if (!pressed) return;

    gp2xzoom_x-=15;
    if (gp2xzoom_x<0) gp2xzoom_x=0;
    GP2X_ZoomSet();
}
static void GP2X_ZoomRight(bool pressed) {
	if (!pressed) return;

    gp2xzoom_x+=15;
    if (gp2xzoom_x>SDL_GetVideoSurface()->w-gp2xzoom_w) gp2xzoom_x=SDL_GetVideoSurface()->w-gp2xzoom_w;
    GP2X_ZoomSet();
}
static void GP2X_ZoomUp(bool pressed) {
	if (!pressed) return;

    gp2xzoom_y-=15;
    if (gp2xzoom_y<0) gp2xzoom_y=0;
    GP2X_ZoomSet();
}
static void GP2X_ZoomDown(bool pressed) {
	if (!pressed) return;

    gp2xzoom_y+=15;
    if (gp2xzoom_y>SDL_GetVideoSurface()->h-gp2xzoom_h) gp2xzoom_y=SDL_GetVideoSurface()->h-gp2xzoom_h;
    GP2X_ZoomSet();
}
static void GP2X_ZoomMore(bool pressed) {
	if (!pressed) return;

	gp2xzoom_pct+=5;
	GP2X_ZoomPct();
}
static void GP2X_ZoomLess(bool pressed) {
	if (!pressed) return;

	gp2xzoom_pct-=5;
	if (gp2xzoom_pct<10) gp2xzoom_pct=10;
	GP2X_ZoomPct();
}

static void GP2X_EmuMouseToggle( bool pressed ) {
	if (!pressed) return;
    gp2x_mouse = !gp2x_mouse;
}

static void GP2X_NoClick( bool pressed ) {
	if (!pressed) return;
    gp2x_noclick = !gp2x_noclick;
}

static void GP2X_LeftClick( bool pressed ) {
    if(pressed){
        Mouse_ButtonPressed(0);
    }
    else {
        Mouse_ButtonReleased(0);
    }
}

static void GP2X_RightClick( bool pressed ) {
    if(pressed){
        Mouse_ButtonPressed(1);
    }
    else {
        Mouse_ButtonReleased(1);
    }
}

static void GP2X_SaveSettings( bool pressed )
{
	if (!pressed) return;
    saveconfig = true;  // Reset this once the file has been written
    GFX_SetTitle( -1, -1, -1 );
}

#if defined (WIN32)
bool GFX_SDLUsingWinDIB(void) {
	return sdl.using_windib;
}
#endif

/* Reset the screen with current values in the sdl structure */
Bitu GFX_GetBestMode(Bitu flags) {
	Bitu testbpp,gotbpp;
	switch (sdl.desktop.want_type) {
	case SCREEN_SURFACE:
check_surface:
		/* Check if we can satisfy the depth it loves */
		if (flags & GFX_LOVE_8) testbpp=8;
		else if (flags & GFX_LOVE_15) testbpp=15;
		else if (flags & GFX_LOVE_16) testbpp=16;
		else if (flags & GFX_LOVE_32) testbpp=32;
		else testbpp=0;
check_gotbpp:
		if (sdl.desktop.fullscreen) gotbpp=SDL_VideoModeOK(640,480,testbpp,SDL_FULLSCREEN|SDL_HWSURFACE|SDL_HWPALETTE);
		else gotbpp=sdl.desktop.bpp;
		/* If we can't get our favorite mode check for another working one */
		switch (gotbpp) {
		case 8:
			if (flags & GFX_CAN_8) flags&=~(GFX_CAN_15|GFX_CAN_16|GFX_CAN_32);
			break;
		case 15:
			if (flags & GFX_CAN_15) flags&=~(GFX_CAN_8|GFX_CAN_16|GFX_CAN_32);
			break;
		case 16:
			if (flags & GFX_CAN_16) flags&=~(GFX_CAN_8|GFX_CAN_15|GFX_CAN_32);
			break;
		case 24:
		case 32:
			if (flags & GFX_CAN_32) flags&=~(GFX_CAN_8|GFX_CAN_15|GFX_CAN_16);
			break;
		}
		flags |= GFX_CAN_RANDOM;
		break;
#if (HAVE_DDRAW_H) && defined(WIN32)
	case SCREEN_SURFACE_DDRAW:
		if (!(flags&(GFX_CAN_15|GFX_CAN_16|GFX_CAN_32))) goto check_surface;
		if (flags & GFX_LOVE_15) testbpp=15;
		else if (flags & GFX_LOVE_16) testbpp=16;
		else if (flags & GFX_LOVE_32) testbpp=32;
		else testbpp=0;
		flags|=GFX_SCALING;
		goto check_gotbpp;
#endif
	case SCREEN_OVERLAY:
		if (flags & GFX_RGBONLY || !(flags&GFX_CAN_32)) goto check_surface;
		flags|=GFX_SCALING;
		flags&=~(GFX_CAN_8|GFX_CAN_15|GFX_CAN_16);
		break;
#if C_OPENGL
	case SCREEN_OPENGL:
		if (flags & GFX_RGBONLY || !(flags&GFX_CAN_32)) goto check_surface;
		flags|=GFX_SCALING;
		flags&=~(GFX_CAN_8|GFX_CAN_15|GFX_CAN_16);
		break;
#endif
	}
	return flags;
}


void GFX_ResetScreen(void) {
	GFX_Stop();
	if (sdl.draw.callback)
		(sdl.draw.callback)( GFX_CallBackReset );
	GFX_Start();
}

static int int_log2 (int val) {
    int log = 0;
    while ((val >>= 1) != 0)
	log++;
    return log;
}


static SDL_Surface * GFX_SetupSurfaceScaled(Bit32u sdl_flags, Bit32u bpp) {
	Bit16u fixedWidth;
	Bit16u fixedHeight;

	if (sdl.desktop.fullscreen) {
		fixedWidth = sdl.desktop.full.fixed ? sdl.desktop.full.width : 0;
		fixedHeight = sdl.desktop.full.fixed ? sdl.desktop.full.height : 0;
		sdl_flags |= SDL_FULLSCREEN|SDL_HWSURFACE;
	} else {
		fixedWidth = sdl.desktop.window.width;
		fixedHeight = sdl.desktop.window.height;
		sdl_flags |= SDL_HWSURFACE;
	}
	if (fixedWidth && fixedHeight) {
		double ratio_w=(double)fixedWidth/(sdl.draw.width*sdl.draw.scalex);
		double ratio_h=(double)fixedHeight/(sdl.draw.height*sdl.draw.scaley);
		if ( ratio_w < ratio_h) {
			sdl.clip.w=fixedWidth;
			sdl.clip.h=(Bit16u)(sdl.draw.height*sdl.draw.scaley*ratio_w);
		} else {
			sdl.clip.w=(Bit16u)(sdl.draw.width*sdl.draw.scalex*ratio_h);
			sdl.clip.h=(Bit16u)fixedHeight;
		}
		if (sdl.desktop.fullscreen)
			sdl.surface = SDL_SetVideoMode(fixedWidth,fixedHeight,bpp,sdl_flags);
		else
			sdl.surface = SDL_SetVideoMode(sdl.clip.w,sdl.clip.h,bpp,sdl_flags);
		if (sdl.surface && sdl.surface->flags & SDL_FULLSCREEN) {
			sdl.clip.x=(Sint16)((sdl.surface->w-sdl.clip.w)/2);
			sdl.clip.y=(Sint16)((sdl.surface->h-sdl.clip.h)/2);
		} else {
			sdl.clip.x = 0;
			sdl.clip.y = 0;
		}
		return sdl.surface;
	} else {
		sdl.clip.x=0;sdl.clip.y=0;
		sdl.clip.w=(Bit16u)(sdl.draw.width*sdl.draw.scalex);
		sdl.clip.h=(Bit16u)(sdl.draw.height*sdl.draw.scaley);
		sdl.surface=SDL_SetVideoMode(sdl.clip.w,sdl.clip.h,bpp,sdl_flags);
		return sdl.surface;
	}
}

Bitu GFX_SetSize(Bitu width,Bitu height,Bitu flags,double scalex,double scaley,GFX_CallBack_t callback) {
	if (sdl.updating)
		GFX_EndUpdate( 0 );

	sdl.draw.width=width;
	sdl.draw.height=height;
	sdl.draw.callback=callback;
	sdl.draw.scalex=scalex;
	sdl.draw.scaley=scaley;

	Bitu bpp=0;
	Bitu retFlags = 0;

	if (sdl.blit.surface) {
		SDL_FreeSurface(sdl.blit.surface);
		sdl.blit.surface=0;
	}
	// GP2X added to eliminate flicker in OSD mode
	if (sdl.blit.buffer) {
		SDL_FreeSurface(sdl.blit.buffer);
		sdl.blit.buffer=0;
	}
	switch (sdl.desktop.want_type) {
	case SCREEN_SURFACE:
dosurface:
		if (flags & GFX_CAN_8) bpp=8;
		if (flags & GFX_CAN_15) bpp=15;
		if (flags & GFX_CAN_16) bpp=16;
		if (flags & GFX_CAN_32) bpp=32;
		sdl.desktop.type=SCREEN_SURFACE;
		sdl.clip.w=width;
		sdl.clip.h=height;
		if (sdl.desktop.fullscreen) {
			if (sdl.desktop.full.fixed) {
				sdl.clip.x=(Sint16)((sdl.desktop.full.width-width)/2);
				sdl.clip.y=(Sint16)((sdl.desktop.full.height-height)/2);
				sdl.surface=SDL_SetVideoMode(sdl.desktop.full.width,sdl.desktop.full.height,bpp,
					SDL_FULLSCREEN | ((flags & GFX_CAN_RANDOM) ? SDL_SWSURFACE : SDL_HWSURFACE) |
					(sdl.desktop.doublebuf ? SDL_DOUBLEBUF|SDL_ASYNCBLIT : 0) | SDL_HWPALETTE);
				if (sdl.surface == NULL) E_Exit("Could not set fullscreen video mode %ix%i-%i: %s",sdl.desktop.full.width,sdl.desktop.full.height,bpp,SDL_GetError());
			} else {
				sdl.clip.x=0;sdl.clip.y=0;
				sdl.surface=SDL_SetVideoMode(width,height,bpp,
					SDL_FULLSCREEN | ((flags & GFX_CAN_RANDOM) ? SDL_SWSURFACE : SDL_HWSURFACE) |
					(sdl.desktop.doublebuf ? SDL_DOUBLEBUF|SDL_ASYNCBLIT  : 0)|SDL_HWPALETTE);
				if (sdl.surface == NULL)
					E_Exit("Could not set fullscreen video mode %ix%i-%i: %s",width,height,bpp,SDL_GetError());
			}
		} else {
			sdl.clip.x=0;sdl.clip.y=0;
			sdl.surface=SDL_SetVideoMode(width,height,bpp,(flags & GFX_CAN_RANDOM) ? SDL_SWSURFACE : SDL_HWSURFACE);
#ifdef WIN32
			if (sdl.surface == NULL) {
				SDL_QuitSubSystem(SDL_INIT_VIDEO);
				if (!sdl.using_windib) {
					LOG_MSG("Failed to create hardware surface.\nRestarting video subsystem with windib enabled.");
					putenv("SDL_VIDEODRIVER=windib");
					sdl.using_windib=true;
				} else {
					LOG_MSG("Failed to create hardware surface.\nRestarting video subsystem with directx enabled.");
					putenv("SDL_VIDEODRIVER=directx");
					sdl.using_windib=false;
				}
				SDL_InitSubSystem(SDL_INIT_VIDEO);
				sdl.surface = SDL_SetVideoMode(width,height,bpp,SDL_HWSURFACE);
			}
#endif
			if (sdl.surface == NULL)
				E_Exit("Could not set windowed video mode %ix%i-%i: %s",width,height,bpp,SDL_GetError());
		}
		if (sdl.surface) {
			switch (sdl.surface->format->BitsPerPixel) {
			case 8:
				retFlags = GFX_CAN_8;
                break;
			case 15:
				retFlags = GFX_CAN_15;
				break;
			case 16:
				retFlags = GFX_CAN_16;
                break;
			case 32:
				retFlags = GFX_CAN_32;
                break;
			}
			if (retFlags && (sdl.surface->flags & SDL_HWSURFACE))
				retFlags |= GFX_HARDWARE;
			//if (retFlags && (sdl.surface->flags & SDL_DOUBLEBUF)) {
            // Force the blit surface to be created for use a buffer surface (GP2X)
				sdl.blit.surface=SDL_CreateRGBSurface(SDL_HWSURFACE,
					sdl.draw.width, sdl.draw.height,
					sdl.surface->format->BitsPerPixel,
					sdl.surface->format->Rmask,
					sdl.surface->format->Gmask,
					sdl.surface->format->Bmask,
				0);
				// GP2X added to eliminate flicker in OSD mode
				sdl.blit.buffer=SDL_CreateRGBSurface(SDL_HWSURFACE,
					sdl.draw.width, sdl.draw.height,
					sdl.surface->format->BitsPerPixel,
					sdl.surface->format->Rmask,
					sdl.surface->format->Gmask,
					sdl.surface->format->Bmask,
				0);
				/* If this one fails be ready for some flickering... */
			//}
		}
		break;
#if (HAVE_DDRAW_H) && defined(WIN32)
	case SCREEN_SURFACE_DDRAW:
		if (flags & GFX_CAN_15) bpp=15;
		if (flags & GFX_CAN_16) bpp=16;
		if (flags & GFX_CAN_32) bpp=32;
		if (!GFX_SetupSurfaceScaled((sdl.desktop.doublebuf && sdl.desktop.fullscreen) ? SDL_DOUBLEBUF : 0,bpp)) goto dosurface;
		sdl.blit.rect.top=sdl.clip.y;
		sdl.blit.rect.left=sdl.clip.x;
		sdl.blit.rect.right=sdl.clip.x+sdl.clip.w;
		sdl.blit.rect.bottom=sdl.clip.y+sdl.clip.h;
		sdl.blit.surface=SDL_CreateRGBSurface(SDL_HWSURFACE,sdl.draw.width,sdl.draw.height,
				sdl.surface->format->BitsPerPixel,
				sdl.surface->format->Rmask,
				sdl.surface->format->Gmask,
				sdl.surface->format->Bmask,
				0);
		if (!sdl.blit.surface || (!sdl.blit.surface->flags&SDL_HWSURFACE)) {
			if (sdl.blit.surface) {
				SDL_FreeSurface(sdl.blit.surface);
				sdl.blit.surface=0;
			}
			LOG_MSG("Failed to create ddraw surface, back to normal surface.");
			goto dosurface;
		}
		switch (sdl.surface->format->BitsPerPixel) {
		case 15:
			retFlags = GFX_CAN_15 | GFX_SCALING | GFX_HARDWARE;
			break;
		case 16:
			retFlags = GFX_CAN_16 | GFX_SCALING | GFX_HARDWARE;
               break;
		case 32:
			retFlags = GFX_CAN_32 | GFX_SCALING | GFX_HARDWARE;
               break;
		}
		sdl.desktop.type=SCREEN_SURFACE_DDRAW;
		break;
#endif
	case SCREEN_OVERLAY:
		if (sdl.overlay) {
			SDL_FreeYUVOverlay(sdl.overlay);
			sdl.overlay=0;
		}
		if (!(flags&GFX_CAN_32) || (flags & GFX_RGBONLY)) goto dosurface;
		if (!GFX_SetupSurfaceScaled(0,0)) goto dosurface;
		sdl.overlay=SDL_CreateYUVOverlay(width*2,height,SDL_UYVY_OVERLAY,sdl.surface);
		if (!sdl.overlay) {
			LOG_MSG("SDL:Failed to create overlay, switching back to surface");
			goto dosurface;
		}
		sdl.desktop.type=SCREEN_OVERLAY;
		retFlags = GFX_CAN_32 | GFX_SCALING | GFX_HARDWARE;
		break;
#if C_OPENGL
	case SCREEN_OPENGL:
	{
		if (sdl.opengl.framebuf) {
#if defined(NVIDIA_PixelDataRange)
			if (sdl.opengl.pixel_data_range) db_glFreeMemoryNV(sdl.opengl.framebuf);
			else
#endif
			free(sdl.opengl.framebuf);
		}
		sdl.opengl.framebuf=0;
		if (!(flags&GFX_CAN_32) || (flags & GFX_RGBONLY)) goto dosurface;
		int texsize=2 << int_log2(width > height ? width : height);
		if (texsize>sdl.opengl.max_texsize) {
			LOG_MSG("SDL:OPENGL:No support for texturesize of %d, falling back to surface",texsize);
			goto dosurface;
		}
		SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );
#if defined (WIN32) && SDL_VERSION_ATLEAST(1, 2, 11)
		SDL_GL_SetAttribute( SDL_GL_SWAP_CONTROL, 0 );
#endif
		GFX_SetupSurfaceScaled(SDL_OPENGL,0);
		if (!sdl.surface || sdl.surface->format->BitsPerPixel<15) {
			LOG_MSG("SDL:OPENGL:Can't open drawing surface, are you running in 16bpp(or higher) mode?");
			goto dosurface;
		}
		/* Create the texture and display list */
#if defined(NVIDIA_PixelDataRange)
		if (sdl.opengl.pixel_data_range) {
			sdl.opengl.framebuf=db_glAllocateMemoryNV(width*height*4,0.0,1.0,1.0);
			glPixelDataRangeNV(GL_WRITE_PIXEL_DATA_RANGE_NV,width*height*4,sdl.opengl.framebuf);
			glEnableClientState(GL_WRITE_PIXEL_DATA_RANGE_NV);
		} else {
#else
		{
#endif
			sdl.opengl.framebuf=malloc(width*height*4);		//32 bit color
		}
		sdl.opengl.pitch=width*4;
		glViewport(sdl.clip.x,sdl.clip.y,sdl.clip.w,sdl.clip.h);
		glMatrixMode (GL_PROJECTION);
		glDeleteTextures(1,&sdl.opengl.texture);
 		glGenTextures(1,&sdl.opengl.texture);
		glBindTexture(GL_TEXTURE_2D,sdl.opengl.texture);
		// No borders
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
		if (sdl.opengl.bilinear) {
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		} else {
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		}

		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, texsize, texsize, 0, GL_BGRA_EXT, GL_UNSIGNED_BYTE, 0);

		glClearColor (0.0, 0.0, 0.0, 1.0);
		glClear(GL_COLOR_BUFFER_BIT);
		SDL_GL_SwapBuffers();
		glClear(GL_COLOR_BUFFER_BIT);
		glShadeModel (GL_FLAT);
		glDisable (GL_DEPTH_TEST);
		glDisable (GL_LIGHTING);
		glDisable(GL_CULL_FACE);
		glEnable(GL_TEXTURE_2D);
		glMatrixMode (GL_MODELVIEW);
		glLoadIdentity ();

		GLfloat tex_width=((GLfloat)(width)/(GLfloat)texsize);
		GLfloat tex_height=((GLfloat)(height)/(GLfloat)texsize);

		if (glIsList(sdl.opengl.displaylist)) glDeleteLists(sdl.opengl.displaylist, 1);
		sdl.opengl.displaylist = glGenLists(1);
		glNewList(sdl.opengl.displaylist, GL_COMPILE);
		glBindTexture(GL_TEXTURE_2D, sdl.opengl.texture);
		glBegin(GL_QUADS);
		// lower left
		glTexCoord2f(0,tex_height); glVertex2f(-1.0f,-1.0f);
		// lower right
		glTexCoord2f(tex_width,tex_height); glVertex2f(1.0f, -1.0f);
		// upper right
		glTexCoord2f(tex_width,0); glVertex2f(1.0f, 1.0f);
		// upper left
		glTexCoord2f(0,0); glVertex2f(-1.0f, 1.0f);
		glEnd();
		glEndList();
		sdl.desktop.type=SCREEN_OPENGL;
		retFlags = GFX_CAN_32 | GFX_SCALING;
#if defined(NVIDIA_PixelDataRange)
		if (sdl.opengl.pixel_data_range)
			retFlags |= GFX_HARDWARE;
#endif
	break;
		}//OPENGL
#endif	//C_OPENGL
	}//CASE
	if (retFlags)
		GFX_Start();
	if (!sdl.mouse.autoenable) SDL_ShowCursor(sdl.mouse.autolock?SDL_DISABLE:SDL_ENABLE);

	SDL_ShowCursor(SDL_DISABLE);    // GP2X get rid the SDL cursor for good!

    if (SDL_GetVideoSurface()->w == 320 && SDL_GetVideoSurface()->h < 240)
    {
	    SDL_GP2X_MiniDisplay(0, 0);
        LOG_MSG("Setting minidisplay");
    }
    else
    {
	    // back to default value
	    SDL_GP2X_MiniDisplay(0, 0); // must reset first

	    SDL_Rect displayarea;
            displayarea.x=displayarea.y=0;
	    displayarea.w=SDL_GetVideoSurface()->w;
        displayarea.h=SDL_GetVideoSurface()->h;
        LOG_MSG("Setting display area %d %d", displayarea.w, displayarea.h);
        SDL_GP2X_Display(&displayarea);
	}

	// reset zoom (GP2X)
	GP2X_ZoomReset();

	return retFlags;
}

void GFX_CaptureMouse(void) {
	sdl.mouse.locked=!sdl.mouse.locked;
	if (sdl.mouse.locked) {
		SDL_WM_GrabInput(SDL_GRAB_ON);
		SDL_ShowCursor(SDL_DISABLE);
	} else {
		SDL_WM_GrabInput(SDL_GRAB_OFF);
		if (sdl.mouse.autoenable || !sdl.mouse.autolock) SDL_ShowCursor(SDL_ENABLE);
	}
    mouselocked=sdl.mouse.locked;
}

bool mouselocked; //Global variable for mapper
static void CaptureMouse(bool pressed) {
	if (!pressed)
		return;
	GFX_CaptureMouse();
}

void GFX_SwitchFullScreen(void) {
	sdl.desktop.fullscreen=!sdl.desktop.fullscreen;
	if (sdl.desktop.fullscreen) {
		if (!sdl.mouse.locked) GFX_CaptureMouse();
	} else {
		if (sdl.mouse.locked) GFX_CaptureMouse();
	}
	GFX_ResetScreen();
}

static void SwitchFullScreen(bool pressed) {
	if (!pressed)
		return;
	GFX_SwitchFullScreen();
}


bool GFX_StartUpdate(Bit8u * & pixels,Bitu & pitch) {
	if (!sdl.active || sdl.updating)
		return false;
	switch (sdl.desktop.type) {
	case SCREEN_SURFACE:
        GP2X_OSD_GFX_Change();
		if (sdl.blit.surface && sdl.blit.buffer && gp2x_osd_bufferstate == OSD_STATE_BUFFER ) {
			if (SDL_MUSTLOCK(sdl.blit.surface) && SDL_LockSurface(sdl.blit.surface) ||
                SDL_MUSTLOCK(sdl.blit.buffer) && SDL_LockSurface(sdl.blit.buffer) )
				return false;
			pixels=(Bit8u *)sdl.blit.surface->pixels;
			pitch=sdl.blit.surface->pitch;
		} else {
			if (SDL_MUSTLOCK(sdl.surface) && SDL_LockSurface(sdl.surface))
				return false;
			pixels=(Bit8u *)sdl.surface->pixels;
			pixels+=sdl.clip.y*sdl.surface->pitch;
			pixels+=sdl.clip.x*sdl.surface->format->BytesPerPixel;
			pitch=sdl.surface->pitch;
		}
		sdl.updating=true;
		return true;
#if (HAVE_DDRAW_H) && defined(WIN32)
	case SCREEN_SURFACE_DDRAW:
		if (SDL_LockSurface(sdl.blit.surface)) {
//			LOG_MSG("SDL Lock failed");
			return false;
		}
		pixels=(Bit8u *)sdl.blit.surface->pixels;
		pitch=sdl.blit.surface->pitch;
		sdl.updating=true;
		return true;
#endif
	case SCREEN_OVERLAY:
		SDL_LockYUVOverlay(sdl.overlay);
		pixels=(Bit8u *)*(sdl.overlay->pixels);
		pitch=*(sdl.overlay->pitches);
		sdl.updating=true;
		return true;
#if C_OPENGL
	case SCREEN_OPENGL:
		pixels=(Bit8u *)sdl.opengl.framebuf;
		pitch=sdl.opengl.pitch;
		sdl.updating=true;
		return true;
#endif
	}
	return false;
}

void GFX_CheckFontSize( void )
{
    if( sdl.surface->w > 320 )
    {
        if( sdl.fontsize != 14 * SDL_GetVideoSurface()->w/320 )
        {
            sdl.fontsize = 14 * SDL_GetVideoSurface()->w/320;
            GP2X_LoadFont( sdl.fontsize );
        }
    }
    else
    {
        if( sdl.fontsize != 14 )
        {
            sdl.fontsize = 14;
            GP2X_LoadFont( sdl.fontsize );
        }
    }
}

// GP2X OSD Draw routine
void GFX_OSD_Debug()
{
	if (sdl.font && gp2x_osd_debug)
	{
		char fpsdisp[300];
		sprintf(fpsdisp, "%s FPS=%3d PGM=%s", sdl.title, sdl.fps, RunningProgram );
        SDL_Color tcolorfg;
        SDL_Color tcolorbg;
		SDL_Surface *tsurf;

		switch (sdl.desktop.type) {
		case SCREEN_SURFACE:
            tcolorfg = basic_colors[fontcolor_front];
            tcolorbg = basic_colors[fontcolor_back];
            tsurf = TTF_RenderText_Shaded(sdl.font, fpsdisp, tcolorfg, tcolorbg );
			if (tsurf) {
				SDL_SetAlpha(tsurf, 0, 255);
                if( fontcolor_back == 15 )
                {
                    SDL_SetColorKey( tsurf, SDL_RLEACCEL | SDL_SRCCOLORKEY, SDL_MapRGB( tsurf->format, basic_colors[15].r
                                                                                                     , basic_colors[15].g
                                                                                                     , basic_colors[15].b ) );
                }
                else
                {
                    SDL_SetColorKey( tsurf, SDL_RLEACCEL | SDL_SRCCOLORKEY, SDL_MapRGB( sdl.surface->format, basic_colors[15].r
                                                                                                           , basic_colors[15].g
                                                                                                           , basic_colors[15].b )  );
                }
				SDL_BlitSurface(tsurf,NULL,sdl.blit.buffer,NULL);
                SDL_FreeSurface(tsurf);
			}
			break;
		}
	}
}

void GFX_OSD_VirtKey()
{
    char key[2] = { 0x00, 0x00 };
    int intCharWidth = 0, intCharHeight = 0;
    SDL_Color tcolorfg;
    SDL_Color tcolorbg;
    SDL_Surface *tsurf;
    SDL_Rect rectPosition;

	if (sdl.font && gp2x_osd_virtkey)
	{
        intCharHeight = TTF_FontHeight( sdl.font );
        rectPosition.x = 0;
        rectPosition.y = sdl.surface->h - intCharHeight;

        // Draw the keyboard
        if( sdl.desktop.type == SCREEN_SURFACE )
        {
            for( int i=VIRT_KEYBOARD_ROW-1; i>=0; i-- )
            {
                for( int j=0; j<VIRT_KEYBOARD_COL; j++ )
                {
                    if( i == gp2x_row_key && ( j == gp2x_col_key ||
                                               j <  VIRT_KEYBOARD_COL/2 && gp2x_low_halfrow == true ||
                                               j >= VIRT_KEYBOARD_COL/2 && gp2x_up_halfrow  == true ||
                                               gp2x_full_row == true ) )
                    {
                        tcolorfg = basic_colors[fontcolor_frontsel];
                        tcolorbg = basic_colors[fontcolor_backsel];
                    }
                    else
                    {
                        tcolorfg = basic_colors[fontcolor_front];
                        tcolorbg = basic_colors[fontcolor_back];
                    }

                    if( gp2x_shift == true && gp2x_cur_kbd == 0 )
                    {
                        if( gp2x_keyboard[gp2x_cur_kbd][i][j] >= 'a' &&
                            gp2x_keyboard[gp2x_cur_kbd][i][j] <= 'z' )
                        {
                            key[0] = gp2x_keyboard[gp2x_cur_kbd][i][j] - 0x20;
                        }
                        else
                        {
                            switch( gp2x_keyboard[gp2x_cur_kbd][i][j] )
                            {
                                case '1':  key[0] = '!'; break;
                                case '2':  key[0] = '@'; break;
                                case '3':  key[0] = '#'; break;
                                case '4':  key[0] = '$'; break;
                                case '5':  key[0] = '%'; break;
                                case '6':  key[0] = '^'; break;
                                case '7':  key[0] = '&'; break;
                                case '8':  key[0] = '*'; break;
                                case '9':  key[0] = '('; break;
                                case '0':  key[0] = ')'; break;
                                case '`':  key[0] = '~'; break;
                                case '-':  key[0] = '_'; break;
                                case '=':  key[0] = '+'; break;
                                case '[':  key[0] = '{'; break;
                                case ']':  key[0] = '}'; break;
                                case ',':  key[0] = '<'; break;
                                case '.':  key[0] = '>'; break;
                                case '/':  key[0] = '?'; break;
                                case ';':  key[0] = ':'; break;
                                case '\\': key[0] = '|'; break;
                                case 0x27: key[0] = '"'; break;
                                default :  key[0] = gp2x_keyboard[gp2x_cur_kbd][i][j]; break;
                            }
                        }
                    }
                    else
                    {
                        key[0] = gp2x_keyboard[gp2x_cur_kbd][i][j];
                    }

                    tsurf = TTF_RenderText_Shaded(sdl.font, &key[0], tcolorfg, tcolorbg );
                    if( tsurf )
                    {
                        SDL_SetAlpha(tsurf, 0, 255);
                        if( fontcolor_back == 15 || fontcolor_backsel == 15 )
                        {
                            SDL_SetColorKey( tsurf, SDL_RLEACCEL | SDL_SRCCOLORKEY, SDL_MapRGB( tsurf->format, basic_colors[15].r
                                                                                                             , basic_colors[15].g
                                                                                                             , basic_colors[15].b ) );
                        }
                        else
                        {
                            SDL_SetColorKey( tsurf, SDL_RLEACCEL | SDL_SRCCOLORKEY, SDL_MapRGB( sdl.surface->format, basic_colors[15].r
                                                                                                                   , basic_colors[15].g
                                                                                                                   , basic_colors[15].b )  );
                        }
                        SDL_BlitSurface(tsurf, NULL, sdl.blit.buffer, &rectPosition);
                        rectPosition.x += tsurf->w;
                        SDL_FreeSurface(tsurf);
                    }
                }
                rectPosition.x = 0;
                rectPosition.y -= intCharHeight;
            }
        }
	}
}

void GFX_EndUpdate( const Bit16u *changedLines ) {
	int ret;
	if (!sdl.updating)
		return;
	sdl.updating=false;

	// Calc FPS for OSD  (GP2X)
	if( gp2x_osd_debug )
	{
        if (sdl.fpstime == 0 || SDL_GetTicks()-sdl.fpstime>=1000) {
            sdl.fps=sdl.curfps;
            sdl.curfps=0;
            sdl.fpstime=SDL_GetTicks();
        }
        sdl.curfps++;
	}

	switch (sdl.desktop.type) {
	case SCREEN_SURFACE:
		if (SDL_MUSTLOCK(sdl.surface)) {
			if (sdl.blit.surface && sdl.blit.buffer && gp2x_osd_bufferstate == OSD_STATE_BUFFER ) {
				SDL_UnlockSurface(sdl.blit.surface);
				SDL_UnlockSurface(sdl.blit.buffer);
                GP2X_OSD_GFX_Change();
                // Copy to a temporary surface to eliminate flickering
				int Blit = SDL_BlitSurface( sdl.blit.surface, NULL, sdl.blit.buffer, &sdl.clip );
				LOG(LOG_MISC,LOG_WARN)("BlitSurface returned %d",Blit);

                GFX_CheckFontSize();
                GFX_OSD_Debug();     // Draw OSD (GP2X)
                GFX_OSD_VirtKey();   // Draw OSD (GP2X)

                // Copy the final surface to the screen
                int BlitBuffer = SDL_BlitSurface( sdl.blit.buffer, NULL, sdl.surface, &sdl.clip );
                LOG(LOG_MISC,LOG_WARN)("BlitBuffer returned %d",BlitBuffer);
			} else {
				SDL_UnlockSurface(sdl.surface);
                GP2X_OSD_GFX_Change();
			}
			SDL_Flip(sdl.surface);
		} else if (changedLines) {
			Bitu y = 0, index = 0, rectCount = 0;
			while (y < sdl.draw.height) {
				if (!(index & 1)) {
					y += changedLines[index];
				} else {
                    SDL_Rect *rect = &sdl.updateRects[rectCount++];
                    rect->x = sdl.clip.x;
                    rect->y = sdl.clip.y + y;
                    rect->w = (Bit16u)sdl.draw.width;
                    rect->h = changedLines[index];
#if 0
                    if (rect->h + rect->y > sdl.surface->h) {
                        LOG_MSG("WTF");
                    }
#endif
                }
				index++;
			}
			if (rectCount)
			{
				SDL_UpdateRects( sdl.surface, rectCount, sdl.updateRects );
			}
		}
		break;
#if (HAVE_DDRAW_H) && defined(WIN32)
	case SCREEN_SURFACE_DDRAW:
		SDL_UnlockSurface(sdl.blit.surface);
		ret=IDirectDrawSurface3_Blt(
			sdl.surface->hwdata->dd_writebuf,&sdl.blit.rect,
			sdl.blit.surface->hwdata->dd_surface,0,
			DDBLT_WAIT, NULL);
		switch (ret) {
		case DD_OK:
			break;
		case DDERR_SURFACELOST:
			IDirectDrawSurface3_Restore(sdl.blit.surface->hwdata->dd_surface);
			IDirectDrawSurface3_Restore(sdl.surface->hwdata->dd_surface);
			break;
		default:
			LOG_MSG("DDRAW:Failed to blit, error %X",ret);
		}
		SDL_Flip(sdl.surface);
		break;
#endif
	case SCREEN_OVERLAY:
		SDL_UnlockYUVOverlay(sdl.overlay);
		SDL_DisplayYUVOverlay(sdl.overlay,&sdl.clip);
		break;
#if C_OPENGL
	case SCREEN_OPENGL:
#if defined(NVIDIA_PixelDataRange)
		if (sdl.opengl.pixel_data_range) {
            glBindTexture(GL_TEXTURE_2D, sdl.opengl.texture);
			glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0,
					sdl.draw.width, sdl.draw.height, GL_BGRA_EXT,
					GL_UNSIGNED_INT_8_8_8_8_REV, sdl.opengl.framebuf);
			glCallList(sdl.opengl.displaylist);
			SDL_GL_SwapBuffers();
		} else
#endif
		if (changedLines) {
			Bitu y = 0, index = 0;
            glBindTexture(GL_TEXTURE_2D, sdl.opengl.texture);
			while (y < sdl.draw.height) {
				if (!(index & 1)) {
					y += changedLines[index];
				} else {
					Bit8u *pixels = (Bit8u *)sdl.opengl.framebuf + y * sdl.opengl.pitch;
					Bitu height = changedLines[index];
					glTexSubImage2D(GL_TEXTURE_2D, 0, 0, y,
						sdl.draw.width, height, GL_BGRA_EXT,
						GL_UNSIGNED_INT_8_8_8_8_REV, pixels );
					y += height;
				}
				index++;
			}
			glCallList(sdl.opengl.displaylist);
			SDL_GL_SwapBuffers();
		}
		break;
#endif

	}
}


void GFX_SetPalette(Bitu start,Bitu count,GFX_PalEntry * entries) {
	/* I should probably not change the GFX_PalEntry :) */
	if (sdl.surface->flags & SDL_HWPALETTE) {
		if (!SDL_SetPalette(sdl.surface,SDL_PHYSPAL,(SDL_Color *)entries,start,count)) {
			E_Exit("SDL:Can't set palette");
		}
	} else {
		if (!SDL_SetPalette(sdl.surface,SDL_LOGPAL,(SDL_Color *)entries,start,count)) {
			E_Exit("SDL:Can't set palette");
		}
	}
}

Bitu GFX_GetRGB(Bit8u red,Bit8u green,Bit8u blue) {
	switch (sdl.desktop.type) {
	case SCREEN_SURFACE:
	case SCREEN_SURFACE_DDRAW:
		return SDL_MapRGB(sdl.surface->format,red,green,blue);
	case SCREEN_OVERLAY:
		{
			Bit8u y =  ( 9797*(red) + 19237*(green) +  3734*(blue) ) >> 15;
			Bit8u u =  (18492*((blue)-(y)) >> 15) + 128;
			Bit8u v =  (23372*((red)-(y)) >> 15) + 128;
#ifdef WORDS_BIGENDIAN
			return (y << 0) | (v << 8) | (y << 16) | (u << 24);
#else
			return (u << 0) | (y << 8) | (v << 16) | (y << 24);
#endif
		}
	case SCREEN_OPENGL:
//		return ((red << 0) | (green << 8) | (blue << 16)) | (255 << 24);
		//USE BGRA
		return ((blue << 0) | (green << 8) | (red << 16)) | (255 << 24);
	}
	return 0;
}

void GFX_Stop() {
	if (sdl.updating)
		GFX_EndUpdate( 0 );
	sdl.active=false;
}

void GFX_Start() {
	sdl.active=true;
}

static void GUI_ShutDown(Section * sec) {
	GFX_Stop();
	if (sdl.draw.callback) (sdl.draw.callback)( GFX_CallBackStop );
	if (sdl.mouse.locked) GFX_CaptureMouse();
	if (sdl.desktop.fullscreen) GFX_SwitchFullScreen();
}

static void KillSwitch(bool pressed) {
	if (!pressed)
		return;
	throw 1;
}

static void SetPriority(PRIORITY_LEVELS level) {

#if C_SET_PRIORITY
// Do nothing if priorties are not the same and not root, else the highest
// priority can not be set as users can only lower priority (not restore it)

	if((sdl.priority.focus != sdl.priority.nofocus ) &&
		(getuid()!=0) ) return;

#endif
	switch (level) {
#ifdef WIN32
	case PRIORITY_LEVEL_PAUSE:	// if DOSBox is paused, assume idle priority
	case PRIORITY_LEVEL_LOWEST:
		SetPriorityClass(GetCurrentProcess(),IDLE_PRIORITY_CLASS);
		break;
	case PRIORITY_LEVEL_LOWER:
		SetPriorityClass(GetCurrentProcess(),BELOW_NORMAL_PRIORITY_CLASS);
		break;
	case PRIORITY_LEVEL_NORMAL:
		SetPriorityClass(GetCurrentProcess(),NORMAL_PRIORITY_CLASS);
		break;
	case PRIORITY_LEVEL_HIGHER:
		SetPriorityClass(GetCurrentProcess(),ABOVE_NORMAL_PRIORITY_CLASS);
		break;
	case PRIORITY_LEVEL_HIGHEST:
		SetPriorityClass(GetCurrentProcess(),HIGH_PRIORITY_CLASS);
		break;
#elif C_SET_PRIORITY
/* Linux use group as dosbox has mulitple threads under linux */
	case PRIORITY_LEVEL_PAUSE:	// if DOSBox is paused, assume idle priority
	case PRIORITY_LEVEL_LOWEST:
		setpriority (PRIO_PGRP, 0,PRIO_MAX);
		break;
	case PRIORITY_LEVEL_LOWER:
		setpriority (PRIO_PGRP, 0,PRIO_MAX-(PRIO_TOTAL/3));
		break;
	case PRIORITY_LEVEL_NORMAL:
		setpriority (PRIO_PGRP, 0,PRIO_MAX-(PRIO_TOTAL/2));
		break;
	case PRIORITY_LEVEL_HIGHER:
		setpriority (PRIO_PGRP, 0,PRIO_MAX-((3*PRIO_TOTAL)/5) );
		break;
	case PRIORITY_LEVEL_HIGHEST:
		setpriority (PRIO_PGRP, 0,PRIO_MAX-((3*PRIO_TOTAL)/4) );
		break;
#endif
	default:
		break;
	}
}

static unsigned char logo[32*32*4]= {
#include "dosbox_logo.h"
};

static void GUI_StartUp(Section * sec) {
	sec->AddDestroyFunction(&GUI_ShutDown);
	Section_prop * section=static_cast<Section_prop *>(sec);
	sdl.active=false;
	sdl.updating=false;

#if !defined(MACOSX)
	/* Set Icon (must be done before any sdl_setvideomode call) */
	/* But don't set it on OS X, as we use a nicer external icon there. */
#if WORDS_BIGENDIAN
	SDL_Surface* logos= SDL_CreateRGBSurfaceFrom((void*)logo,32,32,32,128,0xff000000,0x00ff0000,0x0000ff00,0);
#else
	SDL_Surface* logos= SDL_CreateRGBSurfaceFrom((void*)logo,32,32,32,128,0x000000ff,0x0000ff00,0x00ff0000,0);
#endif
	SDL_WM_SetIcon(logos,NULL);
#endif

	SDL_GP2X_AllowGfxMemory(NULL,0);
	LOG_MSG("Allow GFX mem\n");

    GP2X_MOUSE_SPEED = section->Get_int("gp2x_mouse_speed");        // Get GP2X parameters
    fontcolor_front = section->Get_int("fontcolor_front");
    fontcolor_back = section->Get_int("fontcolor_back");
    fontcolor_frontsel = section->Get_int("fontcolor_frontsel");
    fontcolor_backsel = section->Get_int("fontcolor_backsel");

	sdl.desktop.fullscreen=section->Get_bool("fullscreen");
	sdl.wait_on_error=section->Get_bool("waitonerror");
	const char * priority=section->Get_string("priority");
	if (priority && priority[0]) {
		Bitu next;
		if (!strncasecmp(priority,"lowest",6)) {
			sdl.priority.focus=PRIORITY_LEVEL_LOWEST;next=6;
		} else if (!strncasecmp(priority,"lower",5)) {
			sdl.priority.focus=PRIORITY_LEVEL_LOWER;next=5;
		} else if (!strncasecmp(priority,"normal",6)) {
			sdl.priority.focus=PRIORITY_LEVEL_NORMAL;next=6;
		} else if (!strncasecmp(priority,"higher",6)) {
			sdl.priority.focus=PRIORITY_LEVEL_HIGHER;next=6;
		} else if (!strncasecmp(priority,"highest",7)) {
			sdl.priority.focus=PRIORITY_LEVEL_HIGHEST;next=7;
		} else {
			next=0;sdl.priority.focus=PRIORITY_LEVEL_HIGHER;
		}
		priority=&priority[next];
		if (next && priority[0]==',' && priority[1]) {
			priority++;
			if (!strncasecmp(priority,"lowest",6)) {
				sdl.priority.nofocus=PRIORITY_LEVEL_LOWEST;
			} else if (!strncasecmp(priority,"lower",5)) {
				sdl.priority.nofocus=PRIORITY_LEVEL_LOWER;
			} else if (!strncasecmp(priority,"normal",6)) {
				sdl.priority.nofocus=PRIORITY_LEVEL_NORMAL;
			} else if (!strncasecmp(priority,"higher",6)) {
				sdl.priority.nofocus=PRIORITY_LEVEL_HIGHER;
			} else if (!strncasecmp(priority,"highest",7)) {
				sdl.priority.nofocus=PRIORITY_LEVEL_HIGHEST;
			} else if (!strncasecmp(priority,"pause",5)) {
				/* we only check for pause here, because it makes no sense
				 * for DOSBox to be paused while it has focus
				 */
				sdl.priority.nofocus=PRIORITY_LEVEL_PAUSE;
			} else {
				sdl.priority.nofocus=PRIORITY_LEVEL_NORMAL;
			}
		} else sdl.priority.nofocus=sdl.priority.focus;
	} else {
		sdl.priority.focus=PRIORITY_LEVEL_HIGHER;
		sdl.priority.nofocus=PRIORITY_LEVEL_NORMAL;
	}
	SetPriority(sdl.priority.focus); //Assume focus on startup
	sdl.mouse.locked=false;
	mouselocked=false; //Global for mapper
	sdl.mouse.requestlock=false;
	sdl.desktop.full.fixed=false;
	const char* fullresolution=section->Get_string("fullresolution");
	sdl.desktop.full.width  = 0;
	sdl.desktop.full.height = 0;
	if(fullresolution && *fullresolution) {
		char res[100];
		strncpy( res, fullresolution, sizeof( res ));
		fullresolution = lowcase (res);//so x and X are allowed
		if(strcmp(fullresolution,"original")) {
			sdl.desktop.full.fixed = true;
			char* height = const_cast<char*>(strchr(fullresolution,'x'));
			if(height && * height) {
				*height = 0;
				sdl.desktop.full.height = atoi(height+1);
				sdl.desktop.full.width  = atoi(res);
			}
		}
	}

	sdl.desktop.window.width  = 0;
	sdl.desktop.window.height = 0;
	const char* windowresolution=section->Get_string("windowresolution");
	if(windowresolution && *windowresolution) {
		char res[100];
		strncpy( res,windowresolution, sizeof( res ));
		windowresolution = lowcase (res);//so x and X are allowed
		if(strcmp(windowresolution,"original")) {
			char* height = const_cast<char*>(strchr(windowresolution,'x'));
			if(height && *height) {
				*height = 0;
				sdl.desktop.window.height = atoi(height+1);
				sdl.desktop.window.width  = atoi(res);
			}
		}
	}
	sdl.desktop.doublebuf=section->Get_bool("fulldouble");
	if (!sdl.desktop.full.width) {
#ifdef WIN32
		sdl.desktop.full.width=GetSystemMetrics(SM_CXSCREEN);
#else
		sdl.desktop.full.width=1024;
#endif
	}
	if (!sdl.desktop.full.height) {
#ifdef WIN32
		sdl.desktop.full.height=GetSystemMetrics(SM_CYSCREEN);
#else
		sdl.desktop.full.height=768;
#endif
	}
	sdl.mouse.autoenable=section->Get_bool("autolock");
	if (!sdl.mouse.autoenable) SDL_ShowCursor(SDL_DISABLE);
	sdl.mouse.autolock=false;
	sdl.mouse.sensitivity=section->Get_int("sensitivity");
	const char * output=section->Get_string("output");
	if (!strcasecmp(output,"surface")) {
		sdl.desktop.want_type=SCREEN_SURFACE;
#if (HAVE_DDRAW_H) && defined(WIN32)
	} else if (!strcasecmp(output,"ddraw")) {
		sdl.desktop.want_type=SCREEN_SURFACE_DDRAW;
#endif
	} else if (!strcasecmp(output,"overlay")) {
		sdl.desktop.want_type=SCREEN_OVERLAY;
#if C_OPENGL
	} else if (!strcasecmp(output,"opengl")) {
		sdl.desktop.want_type=SCREEN_OPENGL;
		sdl.opengl.bilinear=true;
	} else if (!strcasecmp(output,"openglnb")) {
		sdl.desktop.want_type=SCREEN_OPENGL;
		sdl.opengl.bilinear=false;
#endif
	} else {
		LOG_MSG("SDL:Unsupported output device %s, switching back to surface",output);
		sdl.desktop.want_type=SCREEN_SURFACE;
	}

	sdl.overlay=0;
#if C_OPENGL
   if(sdl.desktop.want_type==SCREEN_OPENGL){ /* OPENGL is requested */
	sdl.surface=SDL_SetVideoMode(640,400,0,SDL_OPENGL);
	if (sdl.surface == NULL) {
		LOG_MSG("Could not initialize OpenGL, switching back to surface");
		sdl.desktop.want_type=SCREEN_SURFACE;
	} else {
	sdl.opengl.framebuf=0;
	sdl.opengl.texture=0;
	sdl.opengl.displaylist=0;
	glGetIntegerv (GL_MAX_TEXTURE_SIZE, &sdl.opengl.max_texsize);
#if defined(__WIN32__) && defined(NVIDIA_PixelDataRange)
	glPixelDataRangeNV = (PFNGLPIXELDATARANGENVPROC) wglGetProcAddress("glPixelDataRangeNV");
	db_glAllocateMemoryNV = (PFNWGLALLOCATEMEMORYNVPROC) wglGetProcAddress("wglAllocateMemoryNV");
	db_glFreeMemoryNV = (PFNWGLFREEMEMORYNVPROC) wglGetProcAddress("wglFreeMemoryNV");
#endif
	const char * gl_ext = (const char *)glGetString (GL_EXTENSIONS);
	if(gl_ext && *gl_ext){
		sdl.opengl.packed_pixel=(strstr(gl_ext,"EXT_packed_pixels") > 0);
		sdl.opengl.paletted_texture=(strstr(gl_ext,"EXT_paletted_texture") > 0);
#if defined(NVIDIA_PixelDataRange)
		sdl.opengl.pixel_data_range=(strstr(gl_ext,"GL_NV_pixel_data_range") >0 ) &&
			glPixelDataRangeNV && db_glAllocateMemoryNV && db_glFreeMemoryNV;
		sdl.opengl.pixel_data_range = 0;
#endif
    	} else {
		sdl.opengl.packed_pixel=sdl.opengl.paletted_texture=false;
	}
	}
	} /* OPENGL is requested end */

#endif	//OPENGL
	/* Initialize screen for first time */
	sdl.surface=SDL_SetVideoMode(640,400,0,0);
	if (sdl.surface == NULL) E_Exit("Could not initialize video: %s",SDL_GetError());
	sdl.desktop.bpp=sdl.surface->format->BitsPerPixel;
	if (sdl.desktop.bpp==24) {
		LOG_MSG("SDL:You are running in 24 bpp mode, this will slow down things!");
	}
	GFX_Stop();
/* Get some Event handlers */
	MAPPER_AddHandler(KillSwitch,MK_f9,MMOD1,"shutdown","ShutDown");
	MAPPER_AddHandler(CaptureMouse,MK_f10,MMOD1,"capmouse","Cap Mouse");
	MAPPER_AddHandler(SwitchFullScreen,MK_return,MMOD2,"fullscr","Fullscreen");

	// OSD/Zoom Event Handlers (GP2X)
	GP2X_ZoomReset();

	MAPPER_AddHandler(GP2X_Zoom,MK_f3,MMOD1,"g2zoom","zoom GP2X ");
	MAPPER_AddHandler(GP2X_NoZoom,MK_f4,MMOD1,"g2nozoom","no zoom GP2X");
	MAPPER_AddHandler(GP2X_ZoomLeft,MK_f5,MMOD1,"g2zleft","zoom left GP2X");
	MAPPER_AddHandler(GP2X_ZoomRight,MK_f6,MMOD1,"g2zright","zoom right GP2X");
	MAPPER_AddHandler(GP2X_ZoomUp,MK_f7,MMOD1,"g2zup","zoom up GP2X");
	MAPPER_AddHandler(GP2X_ZoomDown,MK_f8,MMOD1,"g2zdown","zoom down GP2X");

	MAPPER_AddHandler(GP2X_ZoomMore,MK_f3,MMOD1,"g2zmore","zoom more GP2X");
	MAPPER_AddHandler(GP2X_ZoomLess,MK_f3,MMOD1,"g2zless","zoom less GP2X");

	MAPPER_AddHandler(GP2X_Debug_Toggle,MK_f3,MMOD1,"g2debug","debug GP2X");
    MAPPER_AddHandler(GP2X_VirtualKeyboard_Toggle,MK_f3,MMOD1,"g2virtkey","toggle keyboard emu GP2X");
	MAPPER_AddHandler(GP2X_EmuMouseToggle,MK_f4,MMOD1,"g2mouse","toggle mouse emu GP2X");
	MAPPER_AddHandler(GP2X_NoClick,MK_f5,MMOD1,"noclick","no clicks GP2X");
    MAPPER_AddHandler(GP2X_LeftClick,MK_f6,MMOD1,"leftclick","send left click GP2X");
	MAPPER_AddHandler(GP2X_RightClick,MK_f7,MMOD1,"rightclick","send right click GP2X");
	MAPPER_AddHandler(GP2X_SaveSettings,MK_f8,MMOD1,"saveconfig","save cycles and fs to file");

#if C_DEBUG
	/* Pause binds with activate-debugger */
#else
	MAPPER_AddHandler(PauseDOSBox,MK_pause,MMOD2,"pause","Pause");
#endif
	/* Get Keyboard state of numlock and capslock */
	SDLMod keystate = SDL_GetModState();
	if(keystate&KMOD_NUM) startup_state_numlock = true;
	if(keystate&KMOD_CAPS) startup_state_capslock = true;
}

void Mouse_AutoLock(bool enable) {
	sdl.mouse.autolock=enable;
	if (sdl.mouse.autoenable) sdl.mouse.requestlock=enable;
	else {
		SDL_ShowCursor(enable?SDL_DISABLE:SDL_ENABLE);
		sdl.mouse.requestlock=false;
	}
}

static void HandleMouseMotion(SDL_MouseMotionEvent * motion) {
	if (sdl.mouse.locked || !sdl.mouse.autoenable)
		Mouse_CursorMoved((float)motion->xrel*sdl.mouse.sensitivity/100.0f,
						  (float)motion->yrel*sdl.mouse.sensitivity/100.0f,
						  (float)(motion->x-sdl.clip.x)/(sdl.clip.w-1)*sdl.mouse.sensitivity/100.0f,
						  (float)(motion->y-sdl.clip.y)/(sdl.clip.h-1)*sdl.mouse.sensitivity/100.0f,
						  sdl.mouse.locked);
}

static void HandleMouseButton(SDL_MouseButtonEvent * button) {
	switch (button->state) {
	case SDL_PRESSED:
		if (sdl.mouse.requestlock && !sdl.mouse.locked) {
			GFX_CaptureMouse();
			// Dont pass klick to mouse handler
			break;
		}
		if (!sdl.mouse.autoenable && sdl.mouse.autolock && button->button == SDL_BUTTON_MIDDLE) {
			GFX_CaptureMouse();
			break;
		}
		switch (button->button) {
		case SDL_BUTTON_LEFT:
			// GP2X hack - move on click
			Mouse_CursorMovedFixed((float)button->x/(float)SDL_GetVideoSurface()->w, (float)button->y/(float)SDL_GetVideoSurface()->h, sdl.mouse.locked);
			if(!gp2x_noclick) Mouse_ButtonPressed(0);
			break;
		case SDL_BUTTON_RIGHT:
			Mouse_ButtonPressed(1);
			break;
		case SDL_BUTTON_MIDDLE:
			Mouse_ButtonPressed(2);
			break;
		}
		break;
	case SDL_RELEASED:
		switch (button->button) {
		case SDL_BUTTON_LEFT:
            Mouse_CursorMovedFixed((float)button->x/(float)SDL_GetVideoSurface()->w, (float)button->y/(float)SDL_GetVideoSurface()->h, sdl.mouse.locked);
			if(!gp2x_noclick) Mouse_ButtonReleased(0);
			break;
		case SDL_BUTTON_RIGHT:
			Mouse_ButtonReleased(1);
			break;
		case SDL_BUTTON_MIDDLE:
			Mouse_ButtonReleased(2);
			break;
		}
		break;
	}
}

static Bit8u laltstate = SDL_KEYUP;
static Bit8u raltstate = SDL_KEYUP;

void GFX_Events() {
	SDL_Event event;
#if defined (REDUCE_JOYSTICK_POLLING)
	static int poll_delay=0;
	int time=GetTicks();
	if (time-poll_delay>20) {
		poll_delay=time;
		if (sdl.num_joysticks>0) SDL_JoystickUpdate();
		MAPPER_UpdateJoysticks();
	}
#endif
	while (SDL_PollEvent(&event)) {
		switch (event.type) {
		case SDL_ACTIVEEVENT:
			if (event.active.state & SDL_APPINPUTFOCUS) {
				if (event.active.gain) {
					if (sdl.desktop.fullscreen && !sdl.mouse.locked)
						GFX_CaptureMouse();
					SetPriority(sdl.priority.focus);
				} else {
					if (sdl.mouse.locked) {
#ifdef WIN32
						if (sdl.desktop.fullscreen) {
							VGA_KillDrawing();
							sdl.desktop.fullscreen=false;
							GFX_ResetScreen();
						}
#endif
						GFX_CaptureMouse();
					}
					SetPriority(sdl.priority.nofocus);
					MAPPER_LosingFocus();
				}
			}

			/* Non-focus priority is set to pause; check to see if we've lost window or input focus
			 * i.e. has the window been minimised or made inactive?
			 */
			if (sdl.priority.nofocus == PRIORITY_LEVEL_PAUSE) {
				if ((event.active.state & (SDL_APPINPUTFOCUS | SDL_APPACTIVE)) && (!event.active.gain)) {
					/* Window has lost focus, pause the emulator.
					 * This is similar to what PauseDOSBox() does, but the exit criteria is different.
					 * Instead of waiting for the user to hit Alt-Break, we wait for the window to
					 * regain window or input focus.
					 */
					bool paused = true;
					SDL_Event ev;

					GFX_SetTitle(-1,-1,true);
					KEYBOARD_ClrBuffer();
					SDL_Delay(500);
					while (SDL_PollEvent(&ev)) {
						// flush event queue.
					}

					while (paused) {
						// WaitEvent waits for an event rather than polling, so CPU usage drops to zero
						SDL_WaitEvent(&ev);

						switch (ev.type) {
						case SDL_QUIT: throw(0); break; // a bit redundant at linux at least as the active events gets before the quit event.
						case SDL_ACTIVEEVENT:     // wait until we get window focus back
							if (ev.active.state & (SDL_APPINPUTFOCUS | SDL_APPACTIVE)) {
								// We've got focus back, so unpause and break out of the loop
								if (ev.active.gain) {
									paused = false;
									GFX_SetTitle(-1,-1,false);
								}

								/* Now poke a "release ALT" command into the keyboard buffer
								 * we have to do this, otherwise ALT will 'stick' and cause
								 * problems with the app running in the DOSBox.
								 */
								KEYBOARD_AddKey(KBD_leftalt, false);
								KEYBOARD_AddKey(KBD_rightalt, false);
							}
							break;
						}
					}
				}
			}
			break;
		case SDL_MOUSEMOTION:
			//HandleMouseMotion(&event.motion);	// GP2X Movement is handled on mouse clicks
			break;
		case SDL_MOUSEBUTTONDOWN:
		case SDL_MOUSEBUTTONUP:
            HandleMouseButton(&event.button);
			break;
		case SDL_VIDEORESIZE:
//			HandleVideoResize(&event.resize);
			break;
		case SDL_QUIT:
			throw(0);
			break;
		case SDL_VIDEOEXPOSE:
			if (sdl.draw.callback) sdl.draw.callback( GFX_CallBackRedraw );
			break;
#ifdef WIN32
		case SDL_KEYDOWN:
		case SDL_KEYUP:
			// ignore event alt+tab
			if (event.key.keysym.sym==SDLK_LALT) laltstate = event.key.type;
			if (event.key.keysym.sym==SDLK_RALT) raltstate = event.key.type;
			if (((event.key.keysym.sym==SDLK_TAB)) &&
				((laltstate==SDL_KEYDOWN) || (raltstate==SDL_KEYDOWN))) break;
#endif

		// GP2X Keyboard/Button Hack BEGIN
		//we're going to do a real ugly hack for the GP2X here...
		//we get the button events for 19 buttons on stick 0
		//so we're going to turn them into key events
		case SDL_JOYBUTTONDOWN:
		case SDL_JOYBUTTONUP:
            button_state = (event.type == SDL_JOYBUTTONDOWN) ? true : false;

            if( gp2x_osd_virtkey == true )
            {
                gp2x_virtkey_press = GP2X_VirtualKeyboard( &event );
            }
            else
            {
                GP2X_ButtonPressed[event.jbutton.button] = button_state;
            }

            if( gp2x_osd_virtkey == false || (gp2x_osd_virtkey == true &&
                                              GP2X_ButtonPressed[event.jbutton.button] == true &&
                                              event.type == SDL_JOYBUTTONUP) )
            {
                if(gp2x_osd_virtkey == true) GP2X_ButtonPressed[event.jbutton.button] = false;

                // Stick/DPAD mouse emulation
                if( GP2X_MOUSE_SPEED >= 0 )
                {
                    switch(event.jbutton.button)
                    {
                        case GP2X_BUTTON_UP:
                            blnStickUp = button_state;
                            break;
                        case GP2X_BUTTON_DOWN:
                            blnStickDown = button_state;
                            break;
                        case GP2X_BUTTON_LEFT:
                            blnStickLeft = button_state;
                            break;
                        case GP2X_BUTTON_RIGHT:
                            blnStickRight = button_state;
                            break;
                        case GP2X_BUTTON_UPLEFT:
                            blnStickUpLeft = button_state;
                            break;
                        case GP2X_BUTTON_UPRIGHT:
                            blnStickUpRight = button_state;
                            break;
                        case GP2X_BUTTON_DOWNLEFT:
                            blnStickDownLeft = button_state;
                            break;
                        case GP2X_BUTTON_DOWNRIGHT:
                            blnStickDownRight = button_state;
                            break;
                    }
                }

                int tmp_button = event.jbutton.button;

                if( gp2x_mouse == false || GP2X_MOUSE_SPEED < 0 || (gp2x_mouse == true && GP2X_MOUSE_SPEED >= 0 && tmp_button > 7) )
                {
                    if (event.type == SDL_JOYBUTTONDOWN)
                    {
                        event.type = SDL_KEYDOWN;
                        event.key.state = SDL_PRESSED;
                    }
                    else
                    {
                        event.type = SDL_KEYUP;
                        event.key.state = SDL_RELEASED;
                    }

                    event.key.keysym.scancode   = 0; // hope no one looks
                    event.key.keysym.sym        = (SDLKey)(SDLK_KP0 + tmp_button); // SDLK_KP0 = 256
                    event.key.keysym.mod        = KMOD_NONE;
                    event.key.keysym.unicode    = 0; // again, hope nobody looks
                }
            }
            else
            {
                if( !gp2x_virtkey_press ) break;
            }
            // dont break and fall through to...
            // GP2X Keyboard/Button Hack END

		default:
			void MAPPER_CheckEvent(SDL_Event * event);
			MAPPER_CheckEvent(&event);
		}
	}

    if( GP2X_MOUSE_SPEED >= 0 && gp2x_mouse == true ) // If the parameter is > 0 then it is turned on (default = -1)
    {
        gp2x_mouse_speed++;

        if( gp2x_mouse_speed >= GP2X_MOUSE_SPEED )
        {
            gp2x_mouse_speed = 0;

            int x_rel = 0, y_rel = 0;
            if( blnStickLeft  || blnStickDownLeft  || blnStickUpLeft   ) x_rel -= gp2x_mouse_pixel;
            if( blnStickRight || blnStickDownRight || blnStickUpRight  ) x_rel += gp2x_mouse_pixel;
            if( blnStickUp    || blnStickUpRight   || blnStickUpLeft   ) y_rel -= gp2x_mouse_pixel;
            if( blnStickDown  || blnStickDownRight || blnStickDownLeft ) y_rel += gp2x_mouse_pixel;
            if( x_rel != 0 || y_rel != 0 )
            {
                Mouse_CursorMoved( x_rel, y_rel, 1.0, 1.0, true );
            }
        }
    }
}

#if defined (WIN32)
static BOOL WINAPI ConsoleEventHandler(DWORD event) {
	switch (event) {
	case CTRL_SHUTDOWN_EVENT:
	case CTRL_LOGOFF_EVENT:
	case CTRL_CLOSE_EVENT:
	case CTRL_BREAK_EVENT:
		raise(SIGTERM);
		return TRUE;
	case CTRL_C_EVENT:
	default: //pass to the next handler
		return FALSE;
	}
}
#endif


/* static variable to show wether there is not a valid stdout.
 * Fixes some bugs when -noconsole is used in a read only directory */
static bool no_stdout = false;

void GFX_ShowMsg(char const* format,...) {
	char buf[512];
	va_list msg;
	va_start(msg,format);
	vsprintf(buf,format,msg);
        strcat(buf,"\n");
	va_end(msg);
	if(!no_stdout) printf(buf);
};

int main(int argc, char* argv[]) {
    printf( "Starting dosbox\n" );
	try {
		CommandLine com_line(argc,argv);
		Config myconf(&com_line);
		control=&myconf;

		/* Can't disable the console with debugger enabled */
#if defined(WIN32) && !(C_DEBUG)
		if (control->cmdline->FindExist("-noconsole")) {
			FreeConsole();
			/* Redirect standard input and standard output */
			if(freopen(STDOUT_FILE, "w", stdout) == NULL)
				no_stdout = true; // No stdout so don't write messages
			freopen(STDERR_FILE, "w", stderr);
			setvbuf(stdout, NULL, _IOLBF, BUFSIZ);	/* Line buffered */
			setbuf(stderr, NULL);					/* No buffering */
		} else {
			if (AllocConsole()) {
				fclose(stdin);
				fclose(stdout);
				fclose(stderr);
				freopen("CONIN$","r",stdin);
				freopen("CONOUT$","w",stdout);
				freopen("CONOUT$","w",stderr);
			}
			SetConsoleTitle("DOSBox Status Window");
		}
#endif  //defined(WIN32) && !(C_DEBUG)
		if (control->cmdline->FindExist("-version") ||
		    control->cmdline->FindExist("--version") ) {
			printf("\nDOSBox version %s, copyright 2002-2007 DOSBox Team.\n\n",VERSION);
			printf("DOSBox is written by the DOSBox Team (See AUTHORS file))\n");
			printf("DOSBox comes with ABSOLUTELY NO WARRANTY.  This is free software,\n");
			printf("and you are welcome to redistribute it under certain conditions;\n");
			printf("please read the COPYING file thoroughly before doing so.\n\n");
			return 0;
		}

#if C_DEBUG
		DEBUG_SetupConsole();
#endif

#if defined(WIN32)
	SetConsoleCtrlHandler((PHANDLER_ROUTINE) ConsoleEventHandler,TRUE);
#endif

#ifdef OS2
        PPIB pib;
        PTIB tib;
        DosGetInfoBlocks(&tib, &pib);
        if (pib->pib_ultype == 2) pib->pib_ultype = 3;
        setbuf(stdout, NULL);
        setbuf(stderr, NULL);
#endif

	/* Display Welcometext in the console */
	LOG_MSG("DOSBox version %s",VERSION);
	LOG_MSG("Copyright 2002-2007 DOSBox Team, published under GNU GPL.");
	LOG_MSG("---");

	/* Init SDL */
	if ( SDL_Init( SDL_INIT_AUDIO|SDL_INIT_VIDEO|SDL_INIT_TIMER|SDL_INIT_CDROM
		|SDL_INIT_NOPARACHUTE
		) < 0 ) E_Exit("Can't init SDL %s",SDL_GetError());

#ifndef DISABLE_JOYSTICK
	//Initialise Joystick seperately. This way we can warn when it fails instead
	//of exiting the application
	if( SDL_InitSubSystem(SDL_INIT_JOYSTICK) < 0 ) LOG_MSG("Failed to init joystick support");

	SDL_JoystickOpen(0); // GP2X Need to force the joystick to open
#endif

	// Init SDL_ttf for OSD (GP2X)
	if ( TTF_Init() < 0 )
		E_Exit("Can't init SDL_ttf %s",TTF_GetError());

    sdl.fontsize = 14;
    GP2X_LoadFont( sdl.fontsize );

	sdl.curfps=sdl.fps=sdl.fpstime=0;

#if defined (WIN32)
#if SDL_VERSION_ATLEAST(1, 2, 10)
		sdl.using_windib=true;
#else
		sdl.using_windib=false;
#endif
		char sdl_drv_name[128];
		if (getenv("SDL_VIDEODRIVER")==NULL) {
			if (SDL_VideoDriverName(sdl_drv_name,128)!=NULL) {
				sdl.using_windib=false;
				if (strcmp(sdl_drv_name,"directx")!=0) {
					SDL_QuitSubSystem(SDL_INIT_VIDEO);
					putenv("SDL_VIDEODRIVER=directx");
					if (SDL_InitSubSystem(SDL_INIT_VIDEO)<0) {
						putenv("SDL_VIDEODRIVER=windib");
						if (SDL_InitSubSystem(SDL_INIT_VIDEO)<0) E_Exit("Can't init SDL Video %s",SDL_GetError());
						sdl.using_windib=true;
					}
				}
			}
		} else {
			char* sdl_videodrv = getenv("SDL_VIDEODRIVER");
			if (strcmp(sdl_videodrv,"directx")==0) sdl.using_windib = false;
			else if (strcmp(sdl_videodrv,"windib")==0) sdl.using_windib = true;
		}
		if (SDL_VideoDriverName(sdl_drv_name,128)!=NULL) {
			if (strcmp(sdl_drv_name,"windib")==0) LOG_MSG("SDL_Init: Starting up with SDL windib video driver.\n          Try to update your video card and directx drivers!");
		}
#endif
		sdl.num_joysticks=SDL_NumJoysticks();
		Section_prop * sdl_sec=control->AddSection_prop("sdl",&GUI_StartUp);
		sdl_sec->AddInitFunction(&MAPPER_StartUp);
		sdl_sec->Add_int("gp2x_mouse_speed", -1);   // Add use stick as mouse position (GP2X)
		sdl_sec->Add_int("fontcolor_front", 0);       // Add option to change default color of vk (GP2X)
        sdl_sec->Add_int("fontcolor_back", 15);   // Add option to change default backgnd color of vk (GP2X)
		sdl_sec->Add_int("fontcolor_frontsel", 1);       // Add option to change default color of vk (GP2X)
        sdl_sec->Add_int("fontcolor_backsel", 5);   // Add option to change default backgnd color of vk (GP2X)
		sdl_sec->Add_bool("fullscreen",false);
		sdl_sec->Add_bool("fulldouble",false);
		sdl_sec->Add_string("fullresolution","original");
		sdl_sec->Add_string("windowresolution","original");
		sdl_sec->Add_string("output","surface");
		sdl_sec->Add_bool("autolock",true);
		sdl_sec->Add_int("sensitivity",100);
		sdl_sec->Add_bool("waitonerror",true);
		sdl_sec->Add_string("priority","higher,normal");
		sdl_sec->Add_string("mapperfile","mapper.txt");
		sdl_sec->Add_bool("usescancodes",true);

		MSG_Add("SDL_CONFIGFILE_HELP",
			"fullscreen -- Start dosbox directly in fullscreen.\n"
			"fulldouble -- Use double buffering in fullscreen.\n"
			"fullresolution -- What resolution to use for fullscreen: original or fixed size (e.g. 1024x768).\n"
			"windowresolution -- Scale the window to this size IF the output device supports hardware scaling.\n"
			"output -- What to use for output: surface,overlay"
#if C_OPENGL
			",opengl,openglnb"
#endif
#if (HAVE_DDRAW_H) && defined(WIN32)
			",ddraw"
#endif
			".\n"
			"autolock -- Mouse will automatically lock, if you click on the screen.\n"
			"sensitiviy -- Mouse sensitivity.\n"
			"waitonerror -- Wait before closing the console if dosbox has an error.\n"
			"priority -- Priority levels for dosbox: lowest,lower,normal,higher,highest,pause (when not focussed).\n"
			"            Second entry behind the comma is for when dosbox is not focused/minimized.\n"
			"mapperfile -- File used to load/save the key/event mappings from.\n"
			"usescancodes -- Avoid usage of symkeys, might not work on all operating systems.\n"
			);
		/* Init all the dosbox subsystems */
		DOSBOX_Init();
		std::string config_file;
		bool parsed_anyconfigfile = false;
		// First parse the configfile in the $HOME directory
		if ((getenv("HOME") != NULL)) {
			config_file = (std::string)getenv("HOME") +
				      (std::string)DEFAULT_CONFIG_FILE;
			if (control->ParseConfigFile(config_file.c_str())) parsed_anyconfigfile = true;
		}
		// Add extra settings from dosbox.conf in the local directory if there is no configfile specified at the commandline
		if (!control->cmdline->FindString("-conf",config_file,true)) config_file="dosbox.conf";
		if (control->ParseConfigFile(config_file.c_str())) parsed_anyconfigfile = true;
		// Add extra settings from additional configfiles at the commandline
		while(control->cmdline->FindString("-conf",config_file,true))
			if (control->ParseConfigFile(config_file.c_str())) parsed_anyconfigfile = true;
		// Give a message if no configfile whatsoever was found.
		if(!parsed_anyconfigfile) LOG_MSG("CONFIG: Using default settings. Create a configfile to change them");

#if (ENVIRON_LINKED)
		control->ParseEnv(environ);
#endif
		/* Init all the sections */
		control->Init();
		/* Some extra SDL Functions */
		if (control->cmdline->FindExist("-fullscreen") || sdl_sec->Get_bool("fullscreen")) {
			if(!sdl.desktop.fullscreen) { //only switch if not allready in fullscreen
				GFX_SwitchFullScreen();
			}
		}

		/* Init the keyMapper */
		MAPPER_Init();
		if (control->cmdline->FindExist("-startmapper")) MAPPER_Run(true);

        // GP2X Load MMU hack
/*        system("/sbin/rmmod mmuhack");
        system("/sbin/insmod ../mmuhack.o");    // Current directory is the game dir, so down one

        int mmufd = open("/dev/mmuhack", O_RDWR);
        if(mmufd < 0)
        {
            printf("MMU hack failed\n");
        }
        else
        {
            printf("MMU hack loaded\n");
            close(mmufd);
        }
*/
		/* Start up main machine */
		control->StartUp();
		/* Shutdown everything */
	} catch (char * error) {
		GFX_ShowMsg("Exit to error: %s",error);
		fflush(NULL);
		if(sdl.wait_on_error) {
			//TODO Maybe look for some way to show message in linux?
#if (C_DEBUG)
			GFX_ShowMsg("Press enter to continue");
			fflush(NULL);
			fgetc(stdin);
#elif defined(WIN32)
			Sleep(5000);
#endif
		}

	}
	catch (int){
		;//nothing pressed killswitch
	}
	catch(...){
		throw;//dunno what happened. rethrow for sdl to catch
	}

	// Close TTF for OSD (GP2X)
	if (sdl.font) TTF_CloseFont(sdl.font);
	TTF_Quit();

	SDL_Quit();//Let's hope sdl will quit as well when it catches an exception

    system("/sbin/rmmod mmuhack"); // GP2X Unload MMU hack

	return 0;
};
