#include "memory_layout.h"


#define WRAPPED(x,y)   ((x)>=0?(x):(x)+(y))
#define MIN(x,y) ((x)<(y)?(x):(y))

typedef struct{
	UINT8	ar;			/* was32 attack rate: AR<<2			*/
	UINT8	dr;			/* was32 decay rate:  DR<<2			*/
	UINT8	rr;			/* was32 release rate:RR<<2			*/
	UINT8	KSR;		/* key scale rate				*/
	UINT8	ksl;		/* keyscale level				*/
	UINT8	ksr;		/* key scale rate: kcode>>KSR	*/
	UINT8	mul;		/* multiple: mul_tab[ML]		*/

	/* Phase Generator */
	UINT32	Cnt;		/* frequency counter			*/
	UINT32	Incr;		/* frequency counter step		*/
	UINT8   FB;			/* feedback shift value			*/
	INT32   *connect;	/* slot output pointer			*/
	INT32   op1_out[2];	/* slot1 output for feedback	*/
	UINT8   CON;		/* connection (algorithm) type	*/

	/* Envelope Generator */
	UINT8	eg_type;	/* percussive/non-percussive mode */
	UINT8	state;		/* phase type					*/
	UINT32	TL;			/* total level: TL << 2			*/
	INT32	TLL;		/* adjusted now TL				*/
	INT32	volume;		/* envelope counter				*/
	UINT16	sl;			/* was32 sustain level: sl_tab[SL]	*/

	UINT32	eg_m_ar;	/* (attack state)				*/
	UINT8	eg_sh_ar;	/* (attack state)				*/
	UINT8	eg_sel_ar;	/* (attack state)				*/
	UINT32	eg_m_dr;	/* (decay state)				*/
	UINT8	eg_sh_dr;	/* (decay state)				*/
	UINT8	eg_sel_dr;	/* (decay state)				*/
	UINT32	eg_m_rr;	/* (release state)				*/
	UINT8	eg_sh_rr;	/* (release state)				*/
	UINT8	eg_sel_rr;	/* (release state)				*/

	UINT8	key;		/* was32 0 = KEY OFF, >0 = KEY ON		*/

	/* LFO */
	UINT8	AMmask;		/* was32 LFO Amplitude Modulation enable mask */
	UINT8	vib;		/* LFO Phase Modulation enable flag (active high)*/

	/* waveform select */
//	UINT8	waveform_number;
// not used
	unsigned int wavetable; // -> UINT8 could help

//unsigned char reserved[128-84];//speedup: pump up the struct size to power of 2
//unsigned char reserved[128-100];//speedup: pump up the struct size to power of 2

} OPL3_SLOT;

typedef struct{
	OPL3_SLOT SLOT[2];

	UINT32	block_fnum;	/* block+fnum					*/
	UINT32	fc;			/* Freq. Increment base			*/
	UINT8	ksl_base;	/* was32 KeyScaleLevel Base step		*/
	UINT8	kcode;		/* key code (for key scaling)	*/

	/*
	   there are 12 2-operator channels which can be combined in pairs
	   to form six 4-operator channel, they are:
		0 and 3,
		1 and 4,
		2 and 5,
		9 and 12,
		10 and 13,
		11 and 14
	*/
	UINT8	extended;	/* set to 1 if this channel forms up a 4op channel with another channel(only used by first of pair of channels, ie 0,1,2 and 9,10,11) */

//unsigned char reserved[512-272];//speedup:pump up the struct size to power of 2

} OPL3_CH;

/* OPL3 state */
typedef struct {
	OPL3_CH	P_CH[18];				/* OPL3 chips have 18 channels	*/

	UINT32	pan[18*4];				/* channels output masks (0xffffffff = enable); 4 masks per one channel */
	UINT32	pan_ctrl_value[18];		/* output control values 1 per one channel (1 value contains 4 masks) */

	UINT32	eg_cnt;					/* global envelope generator counter	*/
	UINT32	eg_timer;				/* global envelope generator counter works at frequency = chipclock/288 (288=8*36) */
	UINT32	eg_timer_add;			/* step of eg_timer						*/
	UINT32	eg_timer_overflow;		/* envelope generator timer overlfows every 1 sample (on real chip) */

	UINT32	*fn_tab;			/* fnumber->increment counter was 1024	*/
//        long long int fn_max;
//#define FNTAB(chip,x)	(UINT32) (((x)*(chip)->fn_max)/1024)

	/* LFO */
	UINT8	lfo_am_depth;
	UINT8	lfo_pm_depth_range;
	UINT32	lfo_am_cnt;
	UINT32	lfo_am_inc;
	UINT32	lfo_pm_cnt;
	UINT32	lfo_pm_inc;

	UINT32	noise_rng;				/* 23 bit noise shift register	*/
	UINT32	noise_p;				/* current noise 'phase'		*/
	UINT32	noise_f;				/* current noise period			*/

	UINT8	OPL3_mode;				/* OPL3 extension enable flag	*/

	UINT8	rhythm;					/* Rhythm mode					*/

	int		T[2];					/* timer counters				*/
	int		TC[2];
	UINT8	st[2];					/* timer enable					*/

	UINT32	address;				/* address register				*/
	UINT8	status;					/* status flag					*/
	UINT8	statusmask;				/* status mask					*/

	UINT8	nts;					/* NTS (note select)			*/

	/* external event callback handlers */

	UINT8 type;						/* chip type					*/
	int clock;						/* master clock  (Hz)			*/
	int rate;						/* sampling rate (Hz)			*/
} OPL3;

#define OPL3_SIZE 4096

/*	TL_TAB_LEN is calculated as:

*	(12+1)=13 - sinus amplitude bits     (Y axis)
*   additional 1: to compensate for calculations of negative part of waveform
*   (if we don't add it then the greatest possible _negative_ value would be -2
*   and we really need -1 for waveform #7)
*	2  - sinus sign bit           (Y axis)
*	TL_RES_LEN - sinus resolution (X axis)
*/
#define TL_RES_LEN	(256)
#define TL_TAB_LEN (13*2*TL_RES_LEN)
static signed int *tl_tab; // was int


#define SIN_BITS		10
#define SIN_LEN			(1<<SIN_BITS)
#define SIN_MASK		(SIN_LEN-1)

/* sin waveform table in 'decibel' scale */
/* four waveforms on OPL2 type chips */
/* sinwave entries */
static unsigned int *sin_tab;  // was int

enum CMESSAGE {
 INIT,
 READ,
 WRITE,
 UPDATE,
 TIMEROVER,
 RESET,
 SHUTDOWN
};

typedef struct {
  int type;
//  int no;
  int data1;
  int data2;
  int data3;
} CoreMessage;

#define MAX_OPL3_CHIPS 2

#define MSG_BUF_SIZE 4*1024
volatile static  CoreMessage* MessageBuffer;
volatile static int* NSubmittedMessages; 
volatile static int* NExecutedMessages; 

/* stereo */
#define SHARED_BUF_SIZE 256
volatile static INT32* SharedBuffer[MAX_OPL3_CHIPS];
volatile static int* BufWritePos[MAX_OPL3_CHIPS];
volatile static int* BufReadPos[MAX_OPL3_CHIPS];

#if (BUILD_YMF262)

static OPL3 *YMF262[MAX_OPL3_CHIPS];	
static int *YMF262NumChips=0;				/* number of chips */

#endif

#define BUFF_BASE_ADDRESS  OPL3_MSG_BASE
#define NSUB_OFFSET		0
#define NEX_OFFSET		(NSUB_OFFSET+sizeof(int))
#define MSG_BUF_OFFSET		(NEX_OFFSET+sizeof(int))
#define BUFPOS_OFFSET		(MSG_BUF_OFFSET+MSG_BUF_SIZE*sizeof(CoreMessage))
#define READPOS_OFFSET		(BUFPOS_OFFSET+MAX_OPL3_CHIPS*sizeof(int))
#define DATA_OFFSET		(READPOS_OFFSET+MAX_OPL3_CHIPS*sizeof(int))
#define TL_TAB_OFFSET		(DATA_OFFSET+MAX_OPL3_CHIPS*SHARED_BUF_SIZE*sizeof(INT32))
#define SIN_TAB_OFFSET		(TL_TAB_OFFSET+TL_TAB_LEN*sizeof(signed int))
#define FNTAB_OFFSET		(SIN_TAB_OFFSET+8*SIN_LEN*sizeof(unsigned int))
#define END_OFFSET		(FNTAB_OFFSET+1024*MAX_OPL3_CHIPS*sizeof(UINT32))

#define DATA_BASE_ADDRESS  OPL3_DAT_BASE
#define NUMCHIP_OFFSET		0
#define OPL_OFFSET		(NUMCHIP_OFFSET+sizeof(int))
#define END_OFFSET2		(OPL_OFFSET+OPL3_SIZE*MAX_OPL3_CHIPS)

static char* SharedData_ptr;
static char* SharedBuff_ptr;
